"""Make protocoll details more robust against wrong user input.

Revision ID: e0985a4462d3
Revises: 668759e73b80
Create Date: 2017-09-18 06:58:53.031849

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'e0985a4462d3'
down_revision = '668759e73b80'


def upgrade():
    op.execute("UPDATE protocoldetail SET duration=0 WHERE duration is null")
    op.alter_column(
        'protocoldetail', 'duration', existing_type=sa.INTEGER(),
        nullable=False, server_default='0')
    op.execute("UPDATE protocoldetail SET message='' WHERE message is null")
    op.alter_column(
        'protocoldetail', 'message', existing_type=sa.TEXT(),
        nullable=False, server_default='')
    op.execute(
        "UPDATE protocoldetail SET responsible='' WHERE responsible is null")
    op.alter_column(
        'protocoldetail', 'responsible', existing_type=sa.VARCHAR(length=2),
        nullable=False, server_default='')


def downgrade():
    op.alter_column(
        'protocoldetail', 'responsible', existing_type=sa.VARCHAR(length=2),
        nullable=True)
    op.alter_column(
        'protocoldetail', 'message', existing_type=sa.TEXT(),
        nullable=True)
    op.alter_column(
        'protocoldetail', 'duration', existing_type=sa.INTEGER(),
        nullable=True)



import pyramid_mailer.interfaces
import pyramid_mailer.mailer


class PrintMailer(pyramid_mailer.mailer.Mailer):
    """
    Dummy mailing instance, printing its mails to console.

    """
    def send(self, message):
        self._print(message)

    def send_immediately(self, message, fail_silently=False):
        self._print(message)

    def send_to_queue(self, message):
        self._print(message)

    def send_sendmail(self, message):
        self._print(message)

    def send_immediately_sendmail(self, message, fail_silently=False):
        self._print(message)

    def _print(self, message):
        if not message.sender:
            message.sender = '<no sender specified>'
        mime_message = message.to_message()
        print(mime_message)


def includeme(config):
    mailer = PrintMailer()
    config.registry.registerUtility(mailer, pyramid_mailer.interfaces.IMailer)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .log import log_with_user, auth_log
from .model import Object, Organization
from datetime import date, timedelta
from fints.client import (
    FinTS3PinTanClient, FinTSClientMode, NeedTANResponse,
    FinTSUnsupportedOperation
)
from sqlalchemy import Column, Integer, String, ForeignKey, DateTime, Boolean
from sqlalchemy import Date
import kontocheck
import pyramid.threadlocal
import sqlalchemy.orm
import transaction


class BankingAccount(Object):
    """ Ein Bankkonto. """

    number = Column(String(10))
    name = Column(String(30))
    last_import = Column(DateTime)


class BookingKind(Object):
    """ Ein Buchungstyp. """

    title = Column(String(50), nullable=False)
    shorttitle = Column(String(10), nullable=True)
    banking_account_id = Column(
        Integer, ForeignKey('bankingaccount.id'), nullable=True)
    banking_account = sqlalchemy.orm.relation(
        'BankingAccount', uselist=False, backref='booking_kinds')


class Booking(Object):
    """ Eine Buchung. """

    banking_account_id = Column(
        Integer, ForeignKey('bankingaccount.id'), nullable=False)
    banking_account = sqlalchemy.orm.relation(
        'BankingAccount', uselist=False, backref='bookings')

    booking_day = Column(Date)
    booking_text = Column(String)
    purpose = Column(String)
    recipient = Column(String)
    iban = Column(String(34))
    bic = Column(String(11))
    value = Column(Integer)

    is_splitted = Column(Boolean, default=False, nullable=False)
    splitted_from_id = Column(Integer, ForeignKey('booking.id'), nullable=True)
    accounting_year = Column(Integer)
    ignore_in_reporting = Column(Boolean, default=False, nullable=False)

    member_id = Column(Integer, ForeignKey('member.id'), nullable=True)
    member = sqlalchemy.orm.relation(
        'Member', uselist=False, backref='bookings')
    kind_id = Column(Integer, ForeignKey('bookingkind.id'), nullable=True)
    kind = sqlalchemy.orm.relation('BookingKind', uselist=False)

    def split(self, new_value):
        value = self.value if self.value is not None else 0
        if value > 0 and new_value < 0:
            return 'Neuer Wert muss positiv sein.'
        if value < 0 and new_value > 0:
            return 'Neuer Wert muss negativ sein.'
        if value > 0 and new_value >= self.value:
            return 'Neuer Wert ist zu groß.'
        if value < 0 and new_value <= self.value:
            return 'Neuer Wert ist zu klein.'
        if new_value == 0:
            return 'Neuer Wert muss größer oder kleiner 0 sein.'
        result = []
        for value in [new_value, self.value - new_value]:
            result.append(Booking.create(
                banking_account=self.banking_account,
                booking_day=self.booking_day,
                booking_text=self.booking_text,
                purpose=self.purpose,
                recipient=self.recipient,
                iban=self.iban,
                bic=self.bic,
                accounting_year=self.accounting_year,
                value=value,
                splitted_from_id=self.id,
                member=self.member,
                kind=self.kind))
        self.is_splitted = True
        return result


class SEPASammler(Object):
    """Ein SEPASammler."""

    pmtinfid = Column(String(100))
    booking_day = Column(Date)
    accounting_year = Column(Integer)
    kind_id = Column(Integer, ForeignKey('bookingkind.id'), nullable=True)
    kind = sqlalchemy.orm.relation('BookingKind', uselist=False)
    is_ueberweisung = Column(Boolean, default=False, nullable=False)


class SEPASammlerEntry(Object):
    """Defines which members are active in a SEPA Sammler."""

    sepa_sammler_id = Column(
        Integer, ForeignKey('sepasammler.id'), nullable=True)
    sepasammler = sqlalchemy.orm.relation(
        'SEPASammler', uselist=False, backref='entries')
    ignore_in_reporting = Column(Boolean, default=False, nullable=False)
    member_id = Column(Integer, ForeignKey('member.id'), nullable=True)
    member = sqlalchemy.orm.relation('Member', uselist=False)
    value = Column(Integer)


class GrundsteuerB(Object):

    parcel_id = Column(
        Integer, ForeignKey('parcel.id'), nullable=False)
    parcel = sqlalchemy.orm.relation('Parcel', uselist=False)
    value = Column(Integer)


class Abwasser(Object):

    parcel_id = Column(
        Integer, ForeignKey('parcel.id'), nullable=False)
    parcel = sqlalchemy.orm.relation('Parcel', uselist=False)
    value = Column(Integer)


def create_fints_client():
    settings = pyramid.threadlocal.get_current_registry().settings
    return FinTS3PinTanClient(
        settings.get('banking.blz'),
        settings.get('banking.kto'),
        settings.get('banking.pin'),
        settings.get('banking.url'),
        mode=FinTSClientMode.INTERACTIVE,
        product_id=settings.get('banking.product_id'))


def ask_for_tan(client, response, user=None):
    log_with_user(
        auth_log.info, user,
        'hat Bank-Import ausgeführt: Fehler ({}).'.format(response.challenge))
    transaction.commit()
    raise NotImplementedError('A TAN is required')
    tan = None
    client.send_tan(response, tan)


def import_transactions_from_fints(user=None):
    kontocheck.lut_load(9)
    client = create_fints_client()
    with client:
        if client.init_tan_response:
            ask_for_tan(client, client.init_tan_response, user)
        accounts = client.get_sepa_accounts()
        if isinstance(accounts, NeedTANResponse):
            accounts = ask_for_tan(client, accounts, user)
        for account in accounts:
            try:
                local_account = (
                    BankingAccount.query()
                    .filter(BankingAccount.number == account.accountnumber)
                    .one())
            except sqlalchemy.orm.exc.NoResultFound:
                continue
            statements = client.get_transactions(
                account, date.today() - timedelta(days=50), date.today())
            while isinstance(statements, NeedTANResponse):
                statements = ask_for_tan(client, statements, user)
            import_transactions(
                [stmt.data for stmt in statements], local_account, user)


def import_transactions(statements, account, user=None):
    count = 0
    for id, statement in enumerate(statements):
        try:
            result = add_transaction(statement, account)
        except Exception as e:
            log_with_user(
                auth_log.info, user,
                'hat Bank-Import ausgeführt: Fehler ({}).'.format(e))
            transaction.commit()
            raise
        if isinstance(result, int):
            count += result
    if count and user:
        log_with_user(
            auth_log.info, user,
            'hat Bank-Import ausgeführt: {} neue Buchungen.'.format(count))
    transaction.commit()


def get_sepa_sammlers(data, account):
    if data['posting_text'] != 'SAMMEL-LS-EINZUG':
        return
    pmtinfid = data['customer_reference'].splitlines()[1]
    try:
        sepasammler = (
            SEPASammler.query()
            .filter(SEPASammler.pmtinfid == pmtinfid)
            .filter(SEPASammler.organization_id == account.organization_id)
            .one())
    except Exception:
        raise ValueError("Keine Sammler zu PmtInfId '{}' gefunden.".format(
            pmtinfid))
    sammlers = (
        SEPASammlerEntry.query()
        .filter(SEPASammlerEntry.sepasammler == sepasammler)
        .filter(
            SEPASammlerEntry.organization_id == account.organization_id)
        .all())
    if not sammlers:
        raise ValueError(
            "Keine Sammler-Einträge zu PmtInfId '{}' gefunden.".format(
                pmtinfid))
    return sammlers


def add_transaction(data, account):
    from sw.allotmentclub import Member
    accounting_year = data['date'].year
    org_title = None
    org = Organization.get(account.organization_id)
    if org:
        org_title = org.title
    if data['currency'] != 'EUR':
        raise ValueError(
            "Couldn't import booking {} because its not in Eur.".format(
                data['purpose']))
    value = int(data['amount'].amount * 10000)
    sammlers = get_sepa_sammlers(data, account)
    booking = Booking.find_or_create(
        organization_id=account.organization_id,
        banking_account=account,
        booking_day=data['date'],
        purpose=data['purpose'],
        recipient=(
            data['applicant_name']
            if data['applicant_name'] is not None else org_title),
        value=value)
    if booking.id:
        return
    booking.accounting_year = accounting_year
    booking.booking_text = data['posting_text']
    booking.iban = data['applicant_iban']
    booking.bic = data['applicant_bin']
    if booking.iban:
        try:
            booking.member = (
                Member.query()
                .filter(Member.iban == booking.iban)
                .filter(Member.organization_id == booking.organization_id)
                .one())
        except Exception:
            pass
    if booking.member:
        try:
            debit_booking = (
                Booking.query()
                .filter(Booking.value == 0 - booking.value)
                .filter(Booking.member == booking.member)
                .filter(Booking.accounting_year == booking.accounting_year)
                .filter(Booking.organization_id == booking.organization_id)
                .one())
        except Exception:
            pass
        else:
            booking.kind = debit_booking.kind
    if booking.booking_text == 'SAMMEL-LS-EINZUG':
        if not booking.recipient:
            booking.recipient = org_title
        sum_ = 0
        for sammler in sammlers:
            sum_ += sammler.value
        if (booking.value - 100) > sum_:
            imported, open_ = booking.split(sum_)
            imported.booking_text = imported.booking_text + ' Verrechnet'
            open_.organization_id = booking.organization_id
            imported.organization_id = booking.organization_id
        sum_ += sammler.value
    return 1



# encoding=utf-8
from __future__ import unicode_literals
from .depot import MAX_FILE_SIZE, get_current_user
from .model import PersonMixin, Member, members_messages_table
from sqlalchemy import Boolean, Column, String, Integer, ForeignKey, Text
from sqlalchemy import DateTime, LargeBinary, Table
from sqlalchemy.orm import relationship
from sw.allotmentclub import Object
import datetime
import sqlalchemy.orm


externals_messages_table = Table(
    'externals_messages', Object.metadata,
    Column('external_id', Integer, ForeignKey('externalrecipient.id')),
    Column('message_id', Integer, ForeignKey('message.id'))
)


class Message(Object):
    """A mail or pdf message."""

    members = relationship("Member", secondary=members_messages_table)
    externals = relationship("ExternalRecipient",
                             secondary=externals_messages_table)
    inbound = Column(Boolean, nullable=False, default=False)
    subject = Column(String(255))
    body = Column(Text)
    user_id = Column(
        Integer, ForeignKey('user.id'), nullable=False)
    user = sqlalchemy.orm.relation('User', uselist=False)
    printed = Column(DateTime)
    sent = Column(DateTime)
    accounting_year = Column(Integer)

    in_reply_to_id = Column(
        Integer, ForeignKey('message.id'), nullable=True)
    in_reply_to = sqlalchemy.orm.relationship('Message', uselist=False)

    attachments = sqlalchemy.orm.relationship(
        'Attachment', uselist=True,
        cascade="all,delete", back_populates='message')

    @property
    def member_ids(self):
        return [m.id for m in self.members]

    @member_ids.setter
    def member_ids(self, value):
        self.members = [Member.get(m) for m in value]

    @property
    def external_ids(self):
        return [e.id for e in self.externals]

    @external_ids.setter
    def external_ids(self, value):
        self.externals = [ExternalRecipient.get(e) for e in value]


class Attachment(Object):
    """An attachment for a message."""

    message_id = Column(
        Integer, ForeignKey('message.id'), nullable=False)
    message = sqlalchemy.orm.relationship(
        'Message', uselist=False, back_populates='attachments')
    filename = Column(String(100), default=u'')
    mimetype = Column(String(100), default=u'')
    size = Column(String(20), default=u'')
    data = Column(LargeBinary(MAX_FILE_SIZE))
    white_page_before = Column(Boolean, default=False, nullable=False)
    date = Column(DateTime, nullable=True, default=datetime.datetime.now)
    user_id = Column(Integer, ForeignKey('user.id'), nullable=True,
                     default=get_current_user)
    user = sqlalchemy.orm.relation(
        'User', uselist=False, backref='attachments')


class SentMessageInfo(Object):
    """Additional info for messages sent via email."""

    message_id = Column(
        Integer, ForeignKey('message.id'), nullable=False)
    message = sqlalchemy.orm.relationship('Message', uselist=False)
    tag = Column(String(150))
    address = Column(String(100))
    status = Column(String(255))


class ExternalRecipient(Object, PersonMixin):
    """An external recipient for messages."""

    title = Column(String(20), default=u'')
    appellation = Column(String(20), default=u'')
    organization = Column(String(100), default='')
    firstname = Column(String(50), default=u'')
    lastname = Column(String(50), default=u'')
    street = Column(String(100), default=u'')
    zip = Column(String(6), default=u'')
    city = Column(String(50), default=u'')
    country = Column(String(50), default=u'')
    email = Column(String(100), default=u'')
    phone = Column(String(50), default=u'')

    messages = relationship("Message", secondary=externals_messages_table)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
import transaction


def test_displays_list_of_log_entries(browser):
    from sw.allotmentclub import Log, User
    User.create(id=2, username='sw', vorname='Sebastian',
                nachname='Wehrmann')
    Log.create(user_id=2, name='user data', level='INFO',
               created=datetime.datetime.now(), msg='Test-Log-Eintrag')
    transaction.commit()

    browser.login()
    browser.open('http://localhost/')
    assert {
        'username': 'sw',
        'firstname': 'Sebastian',
        'lastname': 'Wehrmann',
        'gravatar_url': 'https://www.gravatar.com/avatar/'
                        'd41d8cd98f00b204e9800998ecf8427e',
        'detail': ['Test-Log-Eintrag'],
        'time': 'gerade eben'
    } in browser.json['data']['timeline']


def test_fa_icon_if_system_user(browser):
    from sw.allotmentclub import Log, User
    user = User.find_or_create(username='system')
    Log.create(user=user, name='user data', level='INFO',
               created=datetime.datetime.now(), msg='Test-Log-Eintrag')
    transaction.commit()

    browser.login()
    browser.open('http://localhost/')
    timeline = browser.json['data']['timeline'][0]
    assert timeline['fa_icon'] == 'fa-gear'
    assert 'gravatar_url' not in timeline



# encoding=utf-8
from __future__ import unicode_literals
from sw.allotmentclub import User


def test_send_mail_sends_text_mail(mailer, database):
    from ..letter import send_mail
    user = User.create(username='hans', vorname='Hans', nachname='Wurst')
    send_mail('sw@gocept.com', 'Betreff', 'Dies ist der Inhalt', user)
    assert 1 == len(mailer.outbox)
    message = mailer.outbox[0]
    assert ('Vorstand Leuna-Bungalowgemeinschaft Roter See '
            '<vorstand@roter-see.de>' == message.sender)
    assert ['sw@gocept.com'] == message.recipients
    assert message.body == (
        '\nDies ist der Inhalt\n\nMit freundlichen Grüßen,\n'
        'Im Auftrag des Vorstandes\n\nHans Wurst (None)\n\n')
    assert message.html is None
    assert 'Anschreiben.pdf' == message.attachments[0].filename



# coding:utf-8
from __future__ import unicode_literals
from ..log import auth_log, log_with_user
from .. import User, DashboardData
from datetime import datetime
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.view import view_config, forbidden_view_config
import hashlib
import pyramid.httpexceptions
import pyramid.security
import sw.allotmentclub.browser.app
import sw.allotmentclub.browser.auth
import sw.allotmentclub.browser.base
import sw.allotmentclub.version
import urllib.parse


# Number of failed logins before the account gets locked.
MAX_FAILED_LOGINS = 3

SESSION_TIMEOUT = 'Ihre Sitzung wurde aufgrund zu langer Inaktivität '\
                  'automatisch beendet. Bitte melden Sie sich erneut an.'


@forbidden_view_config(renderer='json')
class Forbidden(sw.allotmentclub.browser.app.HTTPError):

    def __call__(self):
        # respond with HTTPForbidden if user is authenticated
        # use unauthenticated_userid to circumvent privilege check,
        # which would raise another HTTPForbidden
        if pyramid.security.unauthenticated_userid(self.request):
            return super(Forbidden, self).__call__()

        if getattr(self.request, 'session_timeout', False):
            self.request.session.flash(SESSION_TIMEOUT, 'error')

        return {'status': 'error', 'message': 'Zugriff verweigert.'}

    # called from superclass .app.Error
    def log(self):
        auth_log.warning(
            'Ref: %s User %s is not allowed to access %s',
            self.error_ref, self.request.user.username, self.request.url)


def get_next_url(request, default_url):
    parsed_url = urllib.parse.urlparse(request.url)
    query = urllib.parse.parse_qs(parsed_url.query)
    if 'came_from' in query:
        next_url = u'{}://{}{}'.format(
            parsed_url.scheme,
            parsed_url.netloc,
            query['came_from'][0])
    else:
        next_url = default_url
    return next_url


class NetatmoMixin():

    def get_temp_data(self, dest='rotersee'):
        data = (DashboardData.query()
                .filter(getattr(DashboardData, '{}_out_temp'.format(
                    dest)).isnot(None))
                .order_by(DashboardData.date.desc()).first())
        if data is None:
            return dict()
        return dict(temperature=getattr(data, '{}_out_temp'.format(dest)),
                    trend=getattr(data, '{}_out_temp_trend'.format(dest)),
                    date=data.date.strftime('%d.%m.%Y %H:%M Uhr'),
                    sum_rain_24=getattr(data, '{}_rain_24'.format(dest)))


@view_config(route_name='login',
             permission=NO_PERMISSION_REQUIRED,
             renderer='json')
class LoginView(sw.allotmentclub.browser.base.View, NetatmoMixin):

    def get_user_data(self, user):
        gravatar_url = 'https://www.gravatar.com/avatar/%s' % (
            hashlib.md5(user.email.encode('utf-8')).hexdigest())
        return dict(username=user.username,
                    name='%s %s' % (user.vorname, user.nachname),
                    gravatar=gravatar_url)

    def __call__(self):
        result = self.login()
        result['version'] = sw.allotmentclub.version.__version__
        return result

    def login(self):
        if self.request.params.get('to'):
            # timeout occured, redirected from a javascript json request
            self.request.session.flash(SESSION_TIMEOUT, 'error')

        if self.request.session.get('auth.userid'):
            user = User.get(self.request.session['auth.userid'])
            if user is not None:
                return dict(status='success',
                            message='Ihre Sitzung wurde wiederhergestellt.',
                            user=self.get_user_data(user),
                            temp=self.get_temp_data())

        if (self.request.method != 'POST' or
                not self.request.params.get('username')):
            self.update()
            return dict(status='error', message='')

        username = self.request.params.get('username').lower()
        password = self.request.params.get('password')

        user, authenticated = sw.allotmentclub.browser.auth.login(
            username, password)
        if authenticated and user and not user.is_locked:
            self.request.session.invalidate()
            pyramid.security.remember(self.request, user.id)
            msg = u'Anmeldung erfolgreich.'
            log_with_user(
                auth_log.info, user, 'hat sich angemeldet.')
            if user.last_login is not None:
                msg = ('{} Das letzte Mal waren Sie hier {}.'.format(
                    msg, user.last_login.strftime('am %d.%m.%Y um %H:%M Uhr')))
            user.last_login = datetime.now()
            if user.failed_logins:
                if user.failed_logins == 1:
                    tmp_msg = 'wurde {} fehlgeschlagener Anmeldeversuch'
                else:
                    tmp_msg = 'wurden {} fehlgeschlagene Anmeldeversuche'
                msg = ('{} Es {} seit Ihrer letzten Anmeldung '
                       'aufgezeichnet.'.format(
                           msg, tmp_msg.format(user.failed_logins)))
                user.failed_logins = 0
            default_url = self.request.route_url('login')
            next_url = get_next_url(self.request, default_url)
            return dict(status='success', message=msg, next_url=next_url,
                        user=self.get_user_data(user),
                        temp=self.get_temp_data())
        else:
            msg = u'Anmeldung fehlgeschlagen.'
            if user:
                log_with_user(
                    auth_log.info, user,
                    'Anmeldung fehlgeschlagen (falsches Passwort) für %s',
                    user.username)
                user.failed_logins += 1
                if user.failed_logins >= MAX_FAILED_LOGINS:
                    user.lock('MAXLOGINS')
                    log_with_user(
                        auth_log.warning, user,
                        'Accout %s gesperrt: %s fehlerhafte Logins.',
                        user.username, user.failed_logins)
                if user.is_locked:
                    msg = ('{} Ihr Account wurde gesperrt: {}'.format(
                        msg, user.is_locked))
            else:
                auth_log.info(
                    'Anmeldung fehlgeschlagen (falscher Benutzername) für %s',
                    username)

            return dict(status='error', message=msg)


@view_config(route_name='logout', renderer='json',
             permission=NO_PERMISSION_REQUIRED)
def logout_view(request, message=None):
    log_with_user(
        auth_log.info, request.user, 'hat sich abgemeldet.')
    pyramid.security.forget(request)
    # Cannot use delete(), since flash message will not be displayed
    # (Reason not known, but delete isn't in ISession anyway)
    request.session.invalidate()
    if message is None:
        status, message = (
            'success',
            u'Abmeldung abgeschlossen. Sie können sich nun erneut anmelden.')
    else:
        status, message = message
    return dict(status=status, message=message,
                version=sw.allotmentclub.version.__version__)



# encoding=utf8
from __future__ import unicode_literals
from ..conftest import import_members, import_energy_meters, import_bookings
import datetime


def setUp():
    import_members()
    import_energy_meters()
    import_bookings()


def test_meter_has_values_for_each_year(database):
    from ..electricity import ElectricMeter
    setUp()
    meter = ElectricMeter.get(1)
    assert '318992603' == meter.number
    assert 102 == meter.allotment.number
    assert 'Groß' == meter.allotment.member.lastname
    assert 2 == len(meter.energy_values)  # 2013, 2014
    assert 2013 == meter.energy_values[0].year
    assert 6893 == meter.energy_values[0].value


def test_meter_can_be_replaced_by_another_meter(database):
    from ..electricity import ElectricMeter
    import transaction
    setUp()
    old = ElectricMeter.get(3)
    new = old.replace(100, '0815', 0)
    transaction.commit()
    assert old.allotment == new.allotment
    assert new == old.replaced_by
    assert new.get_last_value(new).member == old.get_last_value(old).member
    assert 0 == new.get_last_value(new).usage
    assert 1053 == old.get_last_value(old).usage


def test_value_has_a_member(database):
    from sw.allotmentclub import EnergyValue, Member
    setUp()
    value = EnergyValue.get(1)
    assert Member.get(1) == value.member


def test_value_can_calculate_usage(database):
    from ..electricity import ElectricMeter
    from ..browser.base import format_kwh
    setUp()
    value = ElectricMeter.query().filter(
        ElectricMeter.number == '81112116').one().energy_values[1]
    value.update_data()
    assert 2014 == value.year
    assert '246 kWh' == format_kwh(value.usage)


def test_value_can_calculate_price(database):
    from ..electricity import ElectricMeter
    from ..browser.base import format_eur
    setUp()
    value = ElectricMeter.query().filter(
        ElectricMeter.number == '81112116').one().energy_values[1]
    value.update_data()
    assert 2014 == value.year
    assert '74,29 €' == format_eur(value.price)
    assert '8,17 €' == format_eur(value.fee)
    assert '82,46 €' == format_eur(value.whole_price)


def add_new_meter_values_to_csv(csv):
    lines = csv.splitlines()
    for index, new_value in enumerate(
            [None, b'6893', b'10812', b'24213', b'8412', b'66345', b'5637',
             b'8000', b'972', b'7988', b'8036']):
        # lines[index] = lines[index].decode('utf-8')
        if new_value is not None:
            lines[index] = lines[index] + new_value
    return b'\n'.join(l for l in lines)


def test__EnergyValue__update_data_1(database):
    """Leaves EnergyValue as is if already discounted."""
    from sw.allotmentclub import EnergyValue
    value = EnergyValue.create(
        value=42, usage=1, price=1, fee=1, whole_price=1, to_pay=1,
        advance_pay=1, discounted=True)
    value.update_data()
    assert value.value == 42
    assert value.price == 1
    assert value.usage == 1
    assert value.fee == 1
    assert value.whole_price == 1
    assert value.to_pay == 1
    assert value.advance_pay == 1


def test__EnergyValue__update_data_2(database):
    """Updates bill data of EnergyValue."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    setUp()
    EnergyPrice.create(
        year=2016, price=3020, normal_fee=81700, power_fee=243300)
    meter = ElectricMeter.get(1)
    value = EnergyValue.create(electric_meter=meter, year=2016, value=7132)
    assert value.price is None
    value.update_member()
    value.update_usage()
    value.update_data()
    assert 239 == value.usage
    assert 721780 == value.price
    assert 81700 == value.fee
    assert value.whole_price == value.fee + value.price
    assert value.whole_price == value.to_pay
    assert 267826 == value.advance_pay


def test__EnergyValue__update_data_3(database):
    """Member is updated with owner of meter."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, Member, Allotment
    mueller = Member.create(lastname='Müller')
    meyer = Member.create(lastname='Meyer')
    mueller_allotment = Allotment.create(number='123', member=mueller)
    meter = ElectricMeter.create(allotment=mueller_allotment)
    value = EnergyValue.create(electric_meter=meter, year=2016)
    assert value.member is None
    value.update_member()
    assert value.member is mueller
    # Its possible that someone pays the bill explicitely for a meter
    meter.discount_to = meyer
    value.member = None
    value.update_member()
    assert value.member is meyer


def test__EnergyValue__update_data_4(database):
    """Advance pay is zero if under THRESHOLD."""
    from sw.allotmentclub import EnergyValue, ElectricMeter
    from sw.allotmentclub.electricity import THRESHOLD
    setUp()
    value = (EnergyValue.query().join(ElectricMeter)
             .filter(EnergyValue.year == 2014)
             .filter(ElectricMeter.number == '21292097').one())
    assert value.price <= THRESHOLD
    assert value.usage == 25
    assert value.price == 75500
    assert value.advance_pay == 0


def test__EnergyValue__update_data_5(database):
    """Creates bookings for advanced pay for the next year."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Booking, BookingKind, BankingAccount
    setUp()
    BookingKind.create(title='Energieabschlag I')
    BookingKind.create(title='Energieabschlag II')
    BankingAccount.create(number='3')
    EnergyPrice.create(
        year=2016, price=3020, normal_fee=81700, power_fee=243300)
    meter = ElectricMeter.get(1)
    value = EnergyValue.create(electric_meter=meter, year=2016, value=7132)
    assert Booking.query().filter(Booking.accounting_year == 2017).all() == []
    value.update_member()
    value.update_usage()
    value.update_data()
    ap1, ap2 = Booking.query().filter(Booking.accounting_year == 2017).all()
    assert ap1.value == ap2.value == -267800
    assert ap1.member == ap2.member == value.member
    assert ap1.booking_day == datetime.date(2017, 3, 31)
    assert ap2.booking_day == datetime.date(2017, 6, 30)
    assert ap1.purpose == 'Energieabschlag I für Zähler 318992603'
    assert ap2.purpose == 'Energieabschlag II für Zähler 318992603'


def test__EnergyValue__update_data_6(database):
    """Creates multiple bookings for multiple meters of the same allotment."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Booking, BookingKind, BankingAccount
    setUp()
    BookingKind.create(title='Energieabrechnung')
    BookingKind.create(title='Energieabschlag I')
    BankingAccount.create(number='3')
    EnergyPrice.create(
        year=2016, price=3020, normal_fee=81700, power_fee=243300)
    meter = ElectricMeter.get(1)
    value = EnergyValue.create(electric_meter=meter, year=2016, value=7132)
    meter2 = ElectricMeter.get(2)
    value2 = EnergyValue.create(electric_meter=meter2, year=2016, value=14123)
    meter.allotment = meter2.allotment

    assert Booking.query().filter(Booking.accounting_year == 2017).all() == []
    value.update_member()
    value.update_usage()
    value.update_data()
    value2.update_member()
    value2.update_usage()
    value2.update_data()

    ap1, ap2 = Booking.query().filter(Booking.accounting_year == 2017).all()
    assert ap1.value != ap2.value
    assert ap1.member == ap2.member == value.member
    assert ap1.booking_day == ap2.booking_day == datetime.date(2017, 3, 31)
    assert ap1.purpose == 'Energieabschlag I für Zähler 318992603'
    assert ap2.purpose == 'Energieabschlag I für Zähler 136426011'

    enab1, enab2 = Booking.query().filter(
        Booking.accounting_year == 2016).all()
    assert enab1.value != enab2.value
    assert enab1.member == enab2.member == value.member
    assert enab1.booking_day == enab2.booking_day == datetime.date(2016, 8, 31)
    assert enab1.purpose == 'Energieabrechnung für Zähler 318992603'
    assert enab2.purpose == 'Energieabrechnung für Zähler 136426011'


def test__EnergyValue__usage_1(database):
    """Usage is 0 if value is None."""
    from sw.allotmentclub import EnergyValue, ElectricMeter
    assert 0 == EnergyValue.create(electric_meter=ElectricMeter.create(),
                                   year=2016)._usage


def test__EnergyValue__price_1(database):
    """Price is 0 if no EnergyPrice is defined for the year."""
    from sw.allotmentclub import EnergyValue
    assert 0 == EnergyValue.create(year=2016)._price


def test__EnergyValue__price_2(database):
    """Price is 0 if EnergyPrice has no price."""
    from sw.allotmentclub import EnergyValue, EnergyPrice, ElectricMeter
    from sw.allotmentclub import Member, Allotment
    EnergyPrice.create(year=2016)
    assert 0 == EnergyValue.create(
        electric_meter=ElectricMeter.create(
            allotment=Allotment.create(number='123', member=Member.create())),
        year=2016)._price


def test__EnergyValue__price_3(database):
    """Price is price * usage."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Member, Allotment
    EnergyPrice.create(year=2016, price=2)
    assert 200 == EnergyValue.create(
        electric_meter=ElectricMeter.create(
            allotment=Allotment.create(number='123', member=Member.create())),
        year=2016, usage=100)._price


def test__EnergyValue__fee_1(database):
    """Fee is 0 if no energy price."""
    from sw.allotmentclub import EnergyValue
    assert 0 == EnergyValue.create(year=2016)._fee


def test__EnergyValue__fee_2(database):
    """Fee is 0 for disconnected meters."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Allotment, Member
    EnergyPrice.create(year=2016)
    assert 0 == EnergyValue.create(
        electric_meter=ElectricMeter.create(
            allotment=Allotment.create(number='123', member=Member.create()),
            disconnected=True),
        year=2016)._fee


def test__EnergyValue__fee_3(database):
    """Fee is 0 no normal_fee."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Allotment, Member
    EnergyPrice.create(year=2016, normal_fee=12345)
    assert 0 == EnergyValue.create(
        electric_meter=ElectricMeter.create(
            allotment=Allotment.create(number='123', member=Member.create())),
        year=2016)._fee


def test__EnergyValue__fee_4(database):
    """Fee is 0 no power_fee."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Allotment, Member
    EnergyPrice.create(year=2016, power_fee=12345)
    assert 0 == EnergyValue.create(
        electric_meter=ElectricMeter.create(
            allotment=Allotment.create(number='123', member=Member.create())),
        year=2016)._fee


def test__EnergyValue__fee_5(database):
    """Fee is normal_fee if not electic_power."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Allotment, Member
    EnergyPrice.create(year=2016, normal_fee=1, power_fee=3)
    assert 1 == EnergyValue.create(
        electric_meter=ElectricMeter.create(
            allotment=Allotment.create(number='123', member=Member.create()),
            electric_power=False),
        year=2016)._fee


def test__EnergyValue__fee_6(database):
    """Fee is power_fee if electic_power."""
    from sw.allotmentclub import EnergyValue, ElectricMeter, EnergyPrice
    from sw.allotmentclub import Allotment, Member
    EnergyPrice.create(year=2016, normal_fee=1, power_fee=3)
    assert 3 == EnergyValue.create(
        electric_meter=ElectricMeter.create(
            allotment=Allotment.create(number='123', member=Member.create()),
            electric_power=True),
        year=2016)._fee



"""Add behoerdenbeauftragter

Revision ID: 2a39970710de
Revises: e0985a4462d3
Create Date: 2017-09-20 13:57:49.620450

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '2a39970710de'
down_revision = 'e0985a4462d3'


def upgrade():
    # signature = pkg_resources.resource_stream(
    #     'sw.allotmentclub.signatures', 'ek.png').read()
    # signature = 'data:application/png;base64,{}'.format(
    #     base64.b64encode(signature).decode('utf-8'))
    signature = None
    op.execute("""
INSERT into public.user
    (username, password, unrestricted_access, vorname, nachname, anrede,
     handynummer, ort, position, signature, email, organization_id)
VALUES
    ('ek', '{pw}', false, 'Dr. Eberhard', 'Kietz', 'Herr', '0351 4708959',
     'Dresden', 'Behördenbeauftragter', '{signature}',
     'eberhard.kietz@online.de', 1)
""".format(pw='$2b$12$6gfMaN3CX2soj2YAh5qk2.yCpVAG6I0QNJAEuKSDs18A3XZFLYXDq',
           signature=signature))


def downgrade():
    op.execute("DELETE FROM public.user WHERE username = 'ek'")



# encoding=utf8
from __future__ import unicode_literals
from ..log import user_data_log, log_with_user
from .base import boolean, date_time, to_string, string_agg, format_size
from .base import format_mimetype
from io import StringIO
from pyramid.response import FileIter
from pyramid.view import view_config
from sqlalchemy.sql import func
from sw.allotmentclub import Keylist, Key, KeylistAttachment, Member, Allotment
import collections
import dateutil.parser
import sw.allotmentclub.browser.base


class KeylistQuery(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Schlüsselbuch': 'expand'
    }

    def select(self):
        return (
            self.db.query(
                Keylist.id.label('#'),
                Keylist.subject.label('Schlüsselbuch'),
            )
            .select_from(Keylist))


@view_config(route_name='keylists', renderer='json', permission='view')
class KeylistListView(sw.allotmentclub.browser.base.TableView):

    query_class = KeylistQuery
    available_actions = [
        dict(url='keylist_add', btn_class='btn-success', icon='fa fa-plus',
             title='Neu'),
        dict(url='keylist_edit', btn_class='btn-success', icon='fa fa-pencil',
             title='Bearbeiten'),
        dict(url='keys', btn_class='btn-success', icon='fa fa-list',
             title='Schlüssel'),
        dict(url='keylist_attachment', btn_class='btn-success',
             icon='fa fa-list', title='Anlagen'),
        dict(url='keylist_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen')]


@view_config(route_name='keylist_edit', renderer='json', permission='view')
class KeylistEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Schlüsselbuch bearbeiten'

    @property
    def load_options(self):
        return {
            'subject': {'label': 'Name'}
        }

    @property
    def load_data(self):
        fields = [('subject', self.context.subject)]
        return collections.OrderedDict(fields)

    def save(self, key, value):
        if key == 'rent' and value:
            value = dateutil.parser.parse(value).datetime()
        return super(KeylistEditView, self).save(key, value)


@view_config(route_name='keylist_add', renderer='json', permission='view')
class KeylistAddView(KeylistEditView):

    def __init__(self, context, request):
        context = Keylist.create()
        context.commit()
        super(KeylistAddView, self).__init__(context, request)
        log_with_user(
            user_data_log.info, self.request.user,
            'Schlüsselbuch %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'keylist_edit'


@view_config(route_name='keylist_delete', renderer='json', permission='view')
class KeylistDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = Keylist


class KeyQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Verliehen am': date_time,
        'Verloren': boolean
    }
    data_class = {
        'Seriennummer': 'expand'
    }
    data_hide = {
        'Verloren': 'phone,tablet',
        'Notiz': 'phone,tablet',
        'Mitglied': 'phone',
        'Verliehen am': 'phone',
    }

    def select(self):
        return (
            self.db.query(
                Key.id.label('#'),
                Key.serial.label('Seriennummer'),
                (func.coalesce(to_string(Member.lastname).concat(', ').concat(
                 to_string(Member.firstname).concat(' (')
                 .concat(func.coalesce(string_agg(Allotment.number), 'n/a')))
                    .concat(')'))).label('Mitglied'),
                Key.rent.label('Verliehen am'),
                Key.note.label('Notiz'),
                Key.lost.label('Verloren'))
            .select_from(Key)
            .outerjoin(Member)
            .outerjoin(Allotment)
            .group_by(Key.id, Member.lastname, Member.firstname)
            .filter(Key.keylist == self.context))


@view_config(route_name='keys', renderer='json', permission='view')
class KeysView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Protokolle."""

    query_class = KeyQuery
    default_order_by = 'serial'
    available_actions = [
        dict(url='key_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='key_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='key_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen')]


@view_config(route_name='key_edit', renderer='json', permission='view')
class KeyEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Schlüssel bearbeiten'

    @property
    def load_options(self):
        return {
            'serial': {'label': 'Seriennummer'},
            'member_id': {
                'label': 'Mitglied',
                'source': self.member_source,
                'css_class': 'chosen'
            },
            'rent': {
                'label': 'Verliehen am',
                'css_class': 'datetimepicker'
            },
            'note': {'label': 'Notiz'},
            'lost': {'label': 'Verloren?', 'template': 'form_boolean'}
        }

    @property
    def load_data(self):
        fields = [
            ('serial', self.context.serial),
            ('member_id', self.context.member_id),
            ('rent', self.context.rent),
            ('note', self.context.note),
            ('lost', self.context.lost),
        ]
        return collections.OrderedDict(fields)

    def get_route(self, item, name):
        route = super(KeyEditView, self).get_route(item, name)
        return route.replace('{keylist_id}', str(self.context.keylist.id))


@view_config(route_name='key_add', renderer='json', permission='view')
class KeyAddView(KeyEditView):

    title = 'Schlüssel hinzufügen'

    def __init__(self, context, request):
        context = Key.create(keylist=context)
        context.commit()
        super(KeyAddView, self).__init__(context, request)
        log_with_user(user_data_log.info, self.request.user,
                      'Schlüssel %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'key_edit'


@view_config(route_name='key_delete', renderer='json', permission='view')
class KeyDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = Key

    def log(self):
        if self.deleted is not None:
            deleted = self.context.id
            keylist = self.context.keylist.subject
            log_with_user(user_data_log.info,
                          self.request.user,
                          'Schlüssel %s aus Schlüsselbuch %s gelöscht.',
                          deleted, keylist)


class AttachmentQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Größe': format_size,
        'Dateityp': format_mimetype,
    }
    data_class = {
        'Name': 'expand'
    }
    data_hide = {
        'Dateityp': 'phone,tablet',
        'Größe': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                KeylistAttachment.id.label('#'),
                KeylistAttachment.name.label('Name'),
                KeylistAttachment.mimetype.label('Dateityp'),
                KeylistAttachment.size.label('Größe'))
            .select_from(KeylistAttachment)
            .filter_by(keylist=self.context))


@view_config(route_name='keylist_attachment', renderer='json',
             permission='view')
class KeylistAttachmentsView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Anlagen."""

    query_class = AttachmentQuery
    default_order_by = 'Name'
    available_actions = [
        dict(url='keylist_attachment_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='keylist_attachment_download', btn_class='btn-success',
             icon='fa fa-download', title='Herunterladen')]


@view_config(route_name='keylist_attachment_add', renderer='json',
             permission='view')
class KeylistAttachmentAddView(sw.allotmentclub.browser.base.AddView):

    model = KeylistAttachment

    def parse_file(self):
        file = self.request.params.get('file')
        data = file.file
        data.seek(0)
        data = data.read()
        name = file.filename
        mimetype = file.type
        size = len(data)
        return name, mimetype, size, data

    def log(self, id):
        log_with_user(user_data_log.info,
                      self.request.user,
                      'hat Ablage %s %s.', id, self.action)

    def __call__(self):
        if not self.form_submit():
            return {'status': 'success', 'data': {}}
        name, mimetype, size, data = self.parse_file()
        attachment = self.model.create(
            name=name, mimetype=mimetype, size=size, data=data,
            keylist=self.context)
        attachment.commit()
        log_with_user(user_data_log.info,
                      self.request.user,
                      'hat Anlage %s zu Protokoll %s hinzugefügt.',
                      attachment.id, self.context.id)
        return {'status': 'success'}


@view_config(route_name='keylist_attachment_download',
             permission='view')
class KeylistAttachmentDownloadView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = self.context.mimetype.encode('utf-8')
        response.content_length = int(self.context.size)
        response.content_disposition = (
            b'attachment; filename=' + self.context.name.encode('utf-8'))
        response.app_iter = FileIter(StringIO(self.context.data))
        return response



"""Update transactions to fit new import format.

Revision ID: cc8ba97635c1
Revises: 1755a035e42a
Create Date: 2016-02-02 14:16:40.837958

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'cc8ba97635c1'
down_revision = '1755a035e42a'


def upgrade():
    op.execute("""UPDATE booking SET purpose = 'Versicherungen.
        Unfallversicherung ZIS-289848871106233*Bungalow GemeinschaftDATUM
        16.12.2015, 19.39 UHR1.TAN 804995' WHERE id = 1206;""")
    op.execute("""UPDATE booking SET recipient = 'Gothaer Allgemeine Versiche',
        purpose = '83.661.577869 ABBUCHUNG AM 04.01.2016
        HAFTPFLICHT-VERSICHERUNG H' WHERE id = 1204;""")
    op.execute("""UPDATE booking SET purpose = '41-02-98011329 Gebaeude
        EUR 141,39 41-07-98011328Hausrat EUR 247,61' WHERE id = 1205;""")
    op.execute("""UPDATE booking SET purpose = 'ENTGELT SPKCARD MIT
        CHIPKARTE 5180073894SUSANN TILLACK' WHERE id = 1215""")


def downgrade():
    pass



"""Add iban and bic to member.

Revision ID: 2b1aa8adb13b
Revises: 75aee9c50aa
Create Date: 2015-03-29 19:00:05.393216

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '2b1aa8adb13b'
down_revision = '75aee9c50aa'


def upgrade():
    op.add_column('member',
                  sa.Column('bic', sa.String(length=11), nullable=True))
    op.add_column('member',
                  sa.Column('iban', sa.String(length=34), nullable=True))


def downgrade():
    op.drop_column('member', 'iban')
    op.drop_column('member', 'bic')



# encoding=utf8
from ..log import user_data_log, log_with_user
from .base import date, string_agg, get_selected_year
from .letter import render_pdf
from .protocol import format_markdown
from pyramid.response import FileIter
from pyramid.view import view_config
from sqlalchemy import func
from sqlalchemy.orm import aliased
from sw.allotmentclub import BankingAccount, BookingKind, VALUE_PER_MEMBER
from sw.allotmentclub import Member, Allotment, Parcel, Booking
from sw.allotmentclub import SaleHistory
import collections
import datetime
import dateutil.parser
import sw.allotmentclub.browser.base
import sw.allotmentclub.browser.external


class Query(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Nachname': 'expand'
    }
    data_hide = {
        'Vorname': 'phone,tablet',
        'Flurstück': 'phone,tablet',
        'Mobil': 'phone,tablet',
        'Straße': 'phone,tablet',
        'PLZ': 'phone,tablet',
        'Ort': 'phone,tablet',
        'Telefon': 'phone',
        'E-Mail': 'phone',
        'Bungalow': 'phone',
        'Geburtstag': 'phone',
    }
    formatters = {
        'Geburtstag': date,
    }

    def select(self):
        return (
            self.db.query(
                Member.id.label('#'),
                func.coalesce(
                    string_agg(Allotment.number), 'n/a').label('Bungalow'),
                func.coalesce(
                    string_agg(Parcel.number), 'n/a').label('Flurstück'),
                Member.lastname.label('Nachname'),
                Member.firstname.label('Vorname'),
                Member.street.label('Straße'),
                Member.zip.label('PLZ'),
                Member.city.label('Ort'),
                Member.phone.label('Telefon'),
                Member.mobile.label('Mobil'),
                Member.email.label('E-Mail'),
                Member.birthday.label('Geburtstag'),
            )
            .select_from(Member)
            .outerjoin(Allotment)
            .outerjoin(Parcel)
            .group_by(Member.id, Parcel.leased)
            .filter(Member.leaving_year.is_(None))
        )


@view_config(route_name='member_list', renderer='json', permission='view')
class MemberListView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    default_order_by = 'lastname'
    available_actions = [
        dict(url='direct_debit_letter', btn_class='btn-success',
             icon='fa fa-print', title='Lastschrift drucken'),
        dict(url='become_member_letter', btn_class='btn-success',
             icon='fa fa-print', title='Beitrittserklärung drucken'),
        dict(url='mv_entrance_list', btn_class='btn-success',
             icon='fa fa-print', title='Einlasslisten drucken'),
        dict(url='member_sale', btn_class='btn-success',
             icon='fa fa-money', title='Verkauf'),
        dict(url='member_sale_history', btn_class='btn-success',
             icon='fa fa-money', title='Verkaufhistorie'),
        dict(url='member_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='member_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='membership_fee', btn_class='btn-danger',
             icon='fa fa-money', title='Mitgliedsbeiträge generieren')]


class LeasedQuery(Query):

    def select(self):
        return super(LeasedQuery, self).select().filter(
            Parcel.leased.is_(True))


@view_config(route_name='member_list_leased', renderer='json',
             permission='view')
class MemberListLeasedView(MemberListView):

    query_class = LeasedQuery


class TapWaterQuery(Query):

    def select(self):
        return super(TapWaterQuery, self).select().filter(
            Parcel.tap_water.is_(True))


@view_config(route_name='member_list_tap_water', renderer='json',
             permission='view')
class MemberListTapWaterView(MemberListView):

    query_class = TapWaterQuery


@view_config(route_name='member_edit', renderer='json', permission='view')
class MemberEditView(
        sw.allotmentclub.browser.external.ExternalRecipientEditView):

    title = 'Mitglied bearbeiten'

    @property
    def load_options(self):
        options = super(MemberEditView, self).load_options
        options.update({
            'mobile': {'label': 'Mobil-Telefon'},
            'birthday': {'label': 'Geburtstag'},
            'direct_debit': {'label': 'Lastschrift',
                             'template': 'form_boolean'},
            'direct_debit_account_holder': {
                'label': 'Abweichender Kontoinhaber'},
            'iban': {'label': 'IBAN'},
            'bic': {'label': 'BIC'},
            'direct_debit_date': {
                'label': 'Lastschrift-Datum',
                'css_class': 'datetimepicker'},
            'get_post': {'label': 'Postversand?', 'template': 'form_boolean'},
            'note': {'label': 'Hinweis', 'template': 'form_markdown'},
        })
        return options

    @property
    def load_data(self):
        fields = list(super(MemberEditView, self).load_data.items())
        fields.extend([
            ('phone', self.context.phone),
            ('mobile', self.context.mobile),
            ('birthday', date(self.context.birthday)),
            ('direct_debit', self.context.direct_debit),
            ('direct_debit_account_holder',
                self.context.direct_debit_account_holder),
            ('iban', self.context.iban),
            ('bic', self.context.bic),
            ('direct_debit_date', date(self.context.direct_debit_date)),
            ('get_post', self.context.get_post),
        ])
        return collections.OrderedDict(fields)

    def save(self, key, value):
        if key in ('direct_debit_date', 'birthday'):
            if value:
                value = dateutil.parser.parse(value).date()
            else:
                value = None
        return super(MemberEditView, self).save(key, value)


@view_config(route_name='member_add', renderer='json', permission='view')
class MemberAddView(MemberEditView):

    def __init__(self, context, request):
        context = Member.create()
        context.commit()
        super(MemberAddView, self).__init__(context, request)
        log_with_user(
            user_data_log.info, self.request.user,
            'Mitglied %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'member_edit'


@view_config(route_name='membership_fee', renderer='json', permission='view')
class MembershipFeeView(sw.allotmentclub.browser.base.View):

    def update(self):
        account = (BankingAccount.query()
                   .filter(BankingAccount.number == '1')).one()
        kind = (BookingKind.query()
                .filter(BookingKind.title == 'Mitgliedsbeitrag')).one()
        members = []
        for parcel in Parcel.query().all():
            member = parcel.allotment.member
            if not member:
                continue
            if member.lastname in ['Verein', 'Werkstatt']:
                continue
            members.append(member)
        for member in set(members):
            Booking.create(
                banking_account=account,
                purpose='Mitgliedsbeitrag %s' % get_selected_year(),
                value=0 - VALUE_PER_MEMBER,
                accounting_year=get_selected_year(),
                booking_day=datetime.date(int(get_selected_year()), 3, 31),
                member=member,
                kind=kind)


class MemberLetter(sw.allotmentclub.browser.base.AddView):

    def get_pdf(self):
        subject = self.request.params['subject']
        message = format_markdown(self.request.params['message'])
        attachments = self.request.params.get('attachments')
        member = self.context if self.with_address else None
        return render_pdf(member, subject, message, self.request.user,
                          attachments=attachments)


@view_config(route_name='direct_debit_letter', renderer='json',
             permission='view')
class DirectDebitLetter(sw.allotmentclub.browser.base.View):

    with_address = None
    subject = 'Lastschrifteinzugsermächtigung'
    subsubject = 'Gläubiger-ID: <b>DE42ZZZ00000348413</b>'
    intro = ''
    message = """
<table style="font-size: 10pt;">
  <tbody>
    <tr>
      <td style="width: 30%;">&nbsp;<hr/>Vorname</td>
      <td style="width: 40%">&nbsp;<hr/>Nachname</td>
      <td style="width: 30%;">&nbsp;<hr/>Bungalownummer</td>
    </tr>
    <tr>
      <td style="width: 30%;">&nbsp;<hr/>Straße</td>
      <td style="width: 40%">&nbsp;<hr/>PLZ</td>
      <td style="width: 30%;">&nbsp;<hr/>Ort</td>
    </tr>
    <tr>
      <td style="width: 30%;">&nbsp;<hr/>Telefon</td>
      <td style="width: 40%">&nbsp;<hr/>Mobil</td>
      <td style="width: 30%;">&nbsp;<hr/>E-Mail-Adresse</td>
    </tr>
  </tbody>
</table>

<p>
Der Vorstand der Bungalowgemeinschaft Roter See e.V. wird hiermit widerruflich
ab sofort ermächtigt, die fälligen Mitgliedsbeiträge sowie die
Energiekostenabschläge und den Endabrechnungsbetrag für den Energieverbrauch
mittels Lastschriftverfahren von meinem Konto einzuziehen.
</p>

<table style="font-size: 10pt;">
  <tbody>
    <tr>
      <td style="width: 40%;">&nbsp;<hr/>IBAN</td>
      <td style="width: 20%">&nbsp;<hr/>BIC</td>
      <td style="width: 40%;">&nbsp;<hr/>Name Kreditinstit</td>
    </tr>
    <tr>
      <td style="width: 40%;">&nbsp;<hr/>Ort</td>
      <td style="width: 20%">&nbsp;<hr/>Datum</td>
      <td style="width: 40%;">&nbsp;<hr/>Unterschrift Kontoinhaber</td>
    </tr>
  </tbody>
</table>"""

    def update(self):
        pdf = self.get_pdf()
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = 'application/pdf'
        response.content_disposition = 'attachment; filename=Lastschrift.pdf'
        response.app_iter = FileIter(pdf)
        self.result = response

    def get_pdf(self):
        subject = self.subject
        message = self.intro + self.message
        message = format_markdown(message)
        member = self.context if self.with_address else None
        return render_pdf(member, subject, message, self.request.user,
                          subsubject=self.subsubject)


@view_config(route_name='become_member_letter', renderer='json',
             permission='view')
class BecomeMemberLetter(sw.allotmentclub.browser.base.View):

    with_address = None
    subject = 'Beitrittserklärung'
    intro = ''
    message = """
<table style="font-size: 10pt;">
  <tbody>
    <tr>
      <td style="width: 30%;">&nbsp;<hr/>Vorname</td>
      <td style="width: 40%">&nbsp;<hr/>Nachname</td>
      <td style="width: 30%;">&nbsp;<hr/>Bungalownummer</td>
    </tr>
    <tr>
      <td style="width: 30%;">&nbsp;<hr/>Straße</td>
      <td style="width: 40%">&nbsp;<hr/>PLZ</td>
      <td style="width: 30%;">&nbsp;<hr/>Ort</td>
    </tr>
    <tr>
      <td style="width: 30%;">&nbsp;<hr/>Telefon</td>
      <td style="width: 40%">&nbsp;<hr/>Mobil</td>
      <td style="width: 30%;">&nbsp;<hr/>E-Mail-Adresse</td>
    </tr>
  </tbody>
</table>

<p>
Hiermit erkläre ich meinen Beitritt zum Verein Leuna-Bungalowgemeinschaft
"Roter See" e.V. und erkenne die mir vorgelegte Satzung und die Ordnungen der
Leuna-Bungalowgemeinschaft an.
</p>

<table style="font-size: 10pt;">
  <tbody>
    <tr>
      <td style="width: 40%;">&nbsp;<hr/>Ort</td>
      <td style="width: 20%">&nbsp;<hr/>Datum</td>
      <td style="width: 40%;">&nbsp;<hr/>Unterschrift</td>
    </tr>
  </tbody>
</table>"""

    def update(self):
        pdf = self.get_pdf()
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = 'application/pdf'
        response.content_disposition = 'attachment; filename=Beitritt.pdf'
        response.app_iter = FileIter(pdf)
        self.result = response

    def get_pdf(self):
        subject = self.subject
        message = self.intro + self.message
        message = format_markdown(message)
        member = self.context if self.with_address else None
        return render_pdf(member, subject, message, self.request.user)


@view_config(route_name='member_sale', renderer='json', permission='view')
class MemberSaleView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Grundstücks-Verkauf'

    @property
    def load_options(self):
        return {
            'buyer_id': {
                'source': self.member_source,
                'label': 'Käufer',
                'css_class': 'chosen'
            },
            'transfer_member_due': {
                'label': 'Mitgliedsbeitrag übernehmen',
                'template': 'form_boolean'
            },
            'transfer_advance_pay_one': {
                'label': 'Abschlag I übernehmen',
                'template': 'form_boolean'
            },
            'transfer_advance_pay_two': {
                'label': 'Abschlag II übernehmen',
                'template': 'form_boolean'
            },
            'transfer_energy_bill': {
                'label': 'Energieabrechnung übernehmen',
                'template': 'form_boolean'
            },
            'transfer_assignment_due': {
                'label': 'Arbeitsstunden übernehmen',
                'template': 'form_boolean'
            }
        }

    @property
    def load_data(self):
        fields = [
            ('buyer_id', self.context.id),
            ('transfer_member_due', False),
            ('transfer_advance_pay_one', False),
            ('transfer_advance_pay_two', False),
            ('transfer_energy_bill', False),
            ('transfer_assignment_due', False)]
        return collections.OrderedDict(fields)

    def save(self, key, value):
        year = get_selected_year()
        if key == 'buyer_id':
            buyer = Member.get(value)
            for allotment in self.context.allotments:
                allotment.member = buyer
            SaleHistory.create(seller_id=self.context.id, buyer_id=buyer.id)
            self.context.leaving_year = year
            Booking.create(
                banking_account_id=2,
                booking_day=datetime.date.today(),
                purpose='Aufnahmebeitrag',
                value=-250000,
                member_id=buyer.id,
                kind_id=13,
                accounting_year=year)
        else:
            sale = (SaleHistory.query()
                    .filter(SaleHistory.date == datetime.date.today())
                    .filter(SaleHistory.seller_id == self.context.id)).one()
            buyer = Member.get(sale.buyer_id)
            query = dict(
                transfer_member_due='Mitgliedsbeitrag {}'.format(year),
                transfer_advance_pay_one='Energieabschlag I',
                transfer_advance_pay_two='Energieabschlag II',
                transfer_energy_bill='Energieabrechnung',
                transfer_assignment_due='Arbeitsstunden {}'.format(year)
            )[key]
            for booking in (
                    Booking.query()
                    .filter(Booking.accounting_year == year)
                    .filter(Booking.purpose.ilike('%{}%'.format(query)))
                    .filter(Booking.member == self.context)).all():
                booking.member = buyer
        return True


@view_config(route_name='mv_entrance_list', renderer='json', permission='view')
class MVEntranceListView(sw.allotmentclub.browser.base.XLSXExporterView):

    filename = 'Einlassliste'
    title = 'Liste Einlass MV'

    def query(self):
        return (
            self.db.query(
                Member.lastname.label('Nachname'),
                Member.firstname.label('Vorname'),
                string_agg(Allotment.number).label('Bungalow'),
                Member.email.label('E-Mail-Adresse'),
                Member.direct_debit.label('Lastschrift'),
                Member.id.label('Unterschrift'),
                Member.id.label('Vollmacht'),
            )
            .select_from(Member)
            .join(Allotment)
            .group_by(Member.id)
            .filter(Member.lastname != 'Werkstatt')
            .filter(Member.lastname != 'Verein')
            .order_by(Member.lastname)
        )

    def _data(self, query):
        data = super(MVEntranceListView, self)._data(query)
        for item in data:
            item = list(item)
            item[-1] = item[-2] = ''
            yield item


seller = aliased(Member, name='seller')
buyer = aliased(Member, name='buyer')


class MemberSaleHistoryQuery(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Bungalow': 'expand'
    }
    data_hide = {
        'Käufer': 'phone',
        'Datum': 'phone',
        'Verkäufer': 'phone',
        'Flurstück': 'phone,table',
    }

    def select(self):
        return (
            self.db.query(
                SaleHistory.id.label('#'),
                SaleHistory.date.label('Datum'),
                func.coalesce(
                    string_agg(Allotment.number), 'n/a').label('Bungalow'),
                func.coalesce(
                    string_agg(Parcel.number), 'n/a').label('Flurstück'),
                (seller.lastname + ', ' + seller.firstname).label('Verkäufer'),
                (buyer.lastname + ', ' + buyer.firstname).label('Käufer'),
            )
            .select_from(SaleHistory)
            .join(seller, SaleHistory.seller_id == seller.id)
            .join(buyer, SaleHistory.buyer_id == buyer.id)
            .outerjoin(Allotment)
            .outerjoin(Parcel)
            .group_by(SaleHistory.id, seller.lastname, seller.firstname,
                      buyer.lastname, buyer.firstname)
        )


@view_config(route_name='member_sale_history', renderer='json',
             permission='view')
class MemberSaleHistoryView(sw.allotmentclub.browser.base.TableView):

    query_class = MemberSaleHistoryQuery
    default_order_by = 'date'



"""Change parcel attachments to parcel map

Revision ID: 98db6b34d1a3
Revises: fc8dd9d7a9d7
Create Date: 2019-11-15 07:38:48.934387

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '98db6b34d1a3'
down_revision = 'fc8dd9d7a9d7'


def upgrade():
    op.drop_table('parcelattachment')
    op.add_column(
        'parcel',
        sa.Column(
            'map_data',
            sa.LargeBinary(length=10485760),
            nullable=True
        )
    )
    op.add_column(
        'parcel',
        sa.Column('map_mimetype', sa.String(length=100), nullable=True)
    )
    op.add_column(
        'parcel',
        sa.Column('map_size', sa.String(length=20), nullable=True)
    )


def downgrade():
    op.drop_column('parcel', 'map_size')
    op.drop_column('parcel', 'map_mimetype')
    op.drop_column('parcel', 'map_data')
    op.create_table(
        'parcelattachment',
        sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column(
            'organization_id', sa.INTEGER(), autoincrement=False,
            nullable=False
        ),
        sa.Column(
            'parcel_id', sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.Column(
            'name', sa.VARCHAR(length=100), autoincrement=False, nullable=True
        ),
        sa.Column(
            'mimetype', sa.VARCHAR(length=100), autoincrement=False,
            nullable=True
        ),
        sa.Column(
            'size', sa.VARCHAR(length=20), autoincrement=False, nullable=True
        ),
        sa.Column(
            'data', postgresql.BYTEA(), autoincrement=False, nullable=True
        ),
        sa.ForeignKeyConstraint(
            ['parcel_id'],
            ['parcel.id'],
            name='parcelattachment_parcel_id_fkey'
        ),
        sa.PrimaryKeyConstraint('id', name='parcelattachment_pkey')
    )



# encoding=utf-8
from sw.allotmentclub import Member, Organization, User
from sw.allotmentclub.account import import_transactions_from_fints
from sw.allotmentclub.application import Application
import csv
import datetime
import gocept.logging
import os
import lnetatmo
import pyramid.testing
import pyramid.threadlocal
import requests
import transaction
import xml.etree.ElementTree as ET
import zope.component


ORGANIZATION_ID = 1


def import_members():
    parser = gocept.logging.ArgumentParser(
        description="Import members from excel file.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    parser.add_argument(
        '-i', '--input', default='members.xslx',
        help='Specify the input file. (default: members.xslx)')
    options = parser.parse_args()
    app = Application.from_filename(options.config)
    registry = pyramid.registry.Registry(
        bases=(zope.component.getGlobalSiteManager(),))
    config = pyramid.config.Configurator(
        settings=app.settings, registry=registry)
    config.setup_registry(settings=app.settings)
    request = pyramid.testing.DummyRequest(_registry=registry)
    request.client_addr = '127.0.0.1'
    context = pyramid.threadlocal.manager.get().copy()
    context['request'] = request
    context['registry'] = registry
    pyramid.threadlocal.manager.push(context)
    with open(options.input, 'r') as file:
        reader = csv.reader(file, delimiter=';')
        for id, line in enumerate(reader):
            if id == 0:
                continue
            Member.find_or_create(
                title='',
                appellation='Herr',
                organization_id=2,
                lastname=line[0],
                firstname=line[1],
                street=line[2],
                zip=line[3],
                city=line[4],
                country='Deutschland',
                phone=line[5],
                mobile=line[6],
                email=line[7],
                birthday=line[8])
    transaction.commit()


def import_transactions():
    parser = gocept.logging.ArgumentParser(
        description="Import banking transactions.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    options = parser.parse_args()
    app = Application.from_filename(options.config)
    registry = pyramid.registry.Registry(
        bases=(zope.component.getGlobalSiteManager(),))
    config = pyramid.config.Configurator(
        settings=app.settings, registry=registry)
    config.setup_registry(settings=app.settings)
    request = pyramid.testing.DummyRequest(_registry=registry)
    request.client_addr = '127.0.0.1'
    context = pyramid.threadlocal.manager.get().copy()
    context['request'] = request
    context['registry'] = registry
    pyramid.threadlocal.manager.push(context)
    user = User.by_username('system')
    import_transactions_from_fints(user)


VCF_FILES = {
    1: 'sweh',
    2: 'blauerball'
}


def export_members_vcf():
    parser = gocept.logging.ArgumentParser(
        description="Export members in vcf format.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    options = parser.parse_args()
    Application.from_filename(options.config)
    for organization in Organization.query().all():
        output = ""
        for m in (
                Member.query()
                .filter(Member.organization_id == organization.id)
                .filter(Member.leaving_year.is_(None))):
            output += """BEGIN:VCARD
VERSION:3.0
PRODID:-//Sebastian Wehrmann//sw.allotmentclub//DE
N:{lastname};{firstname};;;
FN:{firstname} {lastname}
ADR;type=HOME;type=pref:;;{street};{city};;{zip};{country}""".format(
                lastname=m.lastname, firstname=m.firstname, street=m.street,
                city=m.city, zip=m.zip, country=m.country)
            if m.email:
                output += """
EMAIL;type=INTERNET;type=HOME;type=pref:{}""".format(m.email)
            if m.mobile:
                output += """
TEL;type=CELL;type=VOICE;type=pref:{}""".format(m.mobile)
            if m.phone:
                output += """
TEL;type=HOME;type=VOICE:{}""".format(m.phone)
            if m.birthday:
                output += """
BDAY:{}""".format(m.birthday.strftime('%Y%m%d'))
            if m.allotments:
                output += """
NOTE:Bungalow: {}""".format('/'.join(str(a.number) for a in m.allotments))
            output += """
ORG:{org}
REV:{rev}""".format(org=organization.title,
                    rev=datetime.datetime.now().isoformat())
            output += """
UID:{uid}
X-RADICALE-NAME:{uid}.vcf
END:VCARD
""".format(uid=m.id)
        path = os.path.join(os.path.expanduser("~"),
                            '.config/radicale/collections',
                            VCF_FILES[organization.id])
        if not os.path.exists(path):
            os.makedirs(path)
        with open(os.path.join(path, 'addressbook.vcf'), 'w') as f:
            f.write(output)


def grab_helios_data(settings):
    payload = {
        "v00003": "de",
        "v01306": "v01306",
        "v00001": "v00001",
        "v00402": settings['helios.pass'],
        "but0": "Anmelden",
    }

    session_requests = requests.session()

    result = session_requests.post(
        'http://10.0.1.64/info.htm',
        data=payload,
        headers=dict(referer='http://10.0.1.64')
    )

    headers = {
        "Host": "10.0.1.64",
        "User-Agent": ("Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:65.0)"
                       " Gecko/20100101 Firefox/65.0"),
        "Accept": "*/*",
        "Accept-Language": "de,en-US;q=0.7,en;q=0.3",
        "Accept-Encoding": "gzip, deflate",
        "Referer": "http://10.0.1.64/anzeig.htm",
        "Referrer-Policy": "no-referrer-when-downgrade",
        "Content-Type": "text/plain;charset=UTF-8",
        "Content-Length": "20",
        "Connection": "keep-alive"}

    result = session_requests.post(
        'http://10.0.1.64/data/werte8.xml',
        headers=headers,
        data="xml=/data/werte8.xml",
    )

    data = dict()
    key = None
    for id, child in enumerate(ET.fromstring(result.text)):
        if id == 0:
            continue
        if key is None:
            key = child
        else:
            data[key.text] = child.text
            key = None

    return data


def grab_netatmo_data(settings):
    result = dict()

    # 1 : Authenticate
    authorization = lnetatmo.ClientAuth(
        settings['netatmo.client.id'],
        settings['netatmo.client.secret'],
        settings['netatmo.user'],
        settings['netatmo.pass'],
        scope="read_station")

    # 2 : Get devices list
    weatherData = lnetatmo.WeatherStationData(authorization)

    # 3 : Access most fresh data directly
    station = weatherData.stationByName('Am Wachtelberg')
    for module in station['modules']:
        if module['type'] == 'NAModule1':
            result['wachtelberg_temp_out_battery'] = module['battery_percent']
            result['wachtelberg_out_temp'] = (
                module['dashboard_data']['Temperature'])
            result['wachtelberg_out_humi'] = (
                module['dashboard_data']['Humidity'])
            temp_trend = module['dashboard_data']['temp_trend']
            temp_trend = ('right' if temp_trend == 'stable' else temp_trend)
            result['wachtelberg_out_temp_trend'] = temp_trend
        elif module['type'] == 'NAModule3':
            result['wachtelberg_rain_battery'] = module['battery_percent']
            result['wachtelberg_rain'] = module['dashboard_data']['Rain']
            result['wachtelberg_rain_1'] = (
                module['dashboard_data']['sum_rain_1'])
            result['wachtelberg_rain_24'] = (
                module['dashboard_data']['sum_rain_24'])
        elif module['type'] == 'NAModule2':
            result['wachtelberg_wind_battery'] = module['battery_percent']
            result['wachtelberg_wind_strength'] = (
                module['dashboard_data']['WindStrength'])
            result['wachtelberg_wind_angle'] = (
                module['dashboard_data']['WindAngle'])
        elif module['type'] == 'NAModule4':
            result['wachtelberg_temp_in_battery'] = module['battery_percent']
            result['wachtelberg_in_temp'] = (
                module['dashboard_data']['Temperature'])
            result['wachtelberg_in_humi'] = (
                module['dashboard_data']['Humidity'])
            result['wachtelberg_in_co2'] = module['dashboard_data']['CO2']

    station = weatherData.stationByName('Roter See')
    for module in station['modules']:
        if module['type'] == 'NAModule1':
            result['rotersee_temp_out_battery'] = module['battery_percent']
            result['rotersee_out_temp'] = (
                module['dashboard_data']['Temperature'])
            result['rotersee_out_humi'] = module['dashboard_data']['Humidity']
            temp_trend = module['dashboard_data']['temp_trend']
            temp_trend = ('right' if temp_trend == 'stable' else temp_trend)
            result['rotersee_out_temp_trend'] = temp_trend
        elif module['type'] == 'NAModule3':
            result['rotersee_rain_battery'] = module['battery_percent']
            result['rotersee_rain'] = module['dashboard_data']['Rain']
            result['rotersee_rain_1'] = module['dashboard_data']['sum_rain_1']
            result['rotersee_rain_24'] = (
                module['dashboard_data']['sum_rain_24'])
    result['rotersee_in_temp'] = station['dashboard_data']['Temperature']
    result['rotersee_in_humi'] = station['dashboard_data']['Humidity']
    result['rotersee_in_co2'] = station['dashboard_data']['CO2']
    return result


def grab_dashboard_data():
    from sw.allotmentclub import DashboardData
    parser = gocept.logging.ArgumentParser(
        description="Grab helios data.")
    parser.add_argument('-c', '--config', default='portal.ini',
                        help='Specify the config file. (default: portal.ini)')
    options = parser.parse_args()
    app = Application.from_filename(options.config)
    registry = pyramid.registry.Registry(
        bases=(zope.component.getGlobalSiteManager(),))
    config = pyramid.config.Configurator(
        settings=app.settings, registry=registry)
    config.setup_registry(settings=app.settings)
    request = pyramid.testing.DummyRequest(_registry=registry)
    request.client_addr = '127.0.0.1'
    context = pyramid.threadlocal.manager.get().copy()
    context['request'] = request
    context['registry'] = registry
    pyramid.threadlocal.manager.push(context)

    heliosdata = grab_helios_data(app.settings)
    try:
        netatmodata = grab_netatmo_data(app.settings)
    except Exception:
        netatmodata = dict()

    dashboarddata = DashboardData.create(
        organization_id=ORGANIZATION_ID,
        luefterstufe=heliosdata['v00102'],
        luefter_percent=heliosdata['v00103'],
        luefter_aussenluft=float(heliosdata['v00104']),
        luefter_zuluft=float(heliosdata['v00105']),
        luefter_fortluft=float(heliosdata['v00106']),
        luefter_abluft=float(heliosdata['v00107']),
        luefter_abluft_feuchte=heliosdata['v02136'])

    for k, v in netatmodata.items():
        setattr(dashboarddata, k, v)

    transaction.commit()



# encoding=utf8
from __future__ import unicode_literals
from .base import format_eur
from pyramid.view import view_config
from sw.allotmentclub import Member, Allotment, Parcel, Abwasser
import sw.allotmentclub.browser.base


class Query(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Nachname': 'expand'
    }

    data_hide = {
        'Vorname': 'phone,tablet',
        'Flurstück': 'phone,tablet',
        'Gebühr': 'phone,tablet',
        'Bungalow': 'phone',
    }

    formatters = {
        'Gebühr': format_eur,
    }

    def select(self):
        return (
            self.db.query(
                Abwasser.id.label('#'),
                Allotment.number.label('Bungalow'),
                Parcel.number.label('Flurstück'),
                Member.lastname.label('Nachname'),
                Member.firstname.label('Vorname'),
                Abwasser.value.label('Gebühr'),
            )
            .select_from(Member)
            .outerjoin(Allotment)
            .outerjoin(Parcel)
            .join(Abwasser)
        )


@view_config(route_name='waste_water', renderer='json', permission='view')
class MemberListView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    default_order_by = 'lastname'



# coding:utf-8

from ..log import auth_log, log_with_user
from ..user import User


def patch():
    from beaker.session import Session
    Session._load_orig = Session.load
    Session.load = load


def load(self):
    """Monkey Patch to generate logs for sessions which timed out"""
    # load session data without timeout check
    import pyramid.threadlocal
    self.namespace = self.namespace_class(
        self.id,
        data_dir=self.data_dir,
        digest_filenames=False,
        **self.namespace_args)
    self.namespace.acquire_read_lock()
    session_data = None
    try:
        try:
            session_data = self.namespace['session']
        except (KeyError, TypeError):
            pass  # new session
    finally:
        self.namespace.release_read_lock()

    # call original method
    self._load_orig()

    # log message if session with user id existed and timeout was reached
    if (session_data is not None and 'auth.userid' in session_data and
            self.was_invalidated):
        user = User.get(session_data['auth.userid'])
        log_with_user(auth_log.info, user,
                      'Session for user %s timed out.', user.username)
        # mark timeout on request to display flash message later on
        request = pyramid.threadlocal.get_current_request()
        request.session_timeout = True



"""Set advance pay to 0 if under threshold.

Revision ID: 1654373c7899
Revises: 51aecbad82b3
Create Date: 2015-03-27 08:15:43.051506

"""

# revision identifiers, used by Alembic.
from alembic import op

revision = '1654373c7899'
down_revision = u'51aecbad82b3'


def upgrade():
    op.execute("""UPDATE energyvalue SET advance_pay = 0 WHERE year = 2014
                  and advance_pay < 100000;""")
    op.execute("""UPDATE booking SET value = 0 WHERE
                  purpose = 'Energieabschlag I' and value > -100000;""")
    op.execute("""UPDATE booking SET value = 0 WHERE
                  purpose = 'Energieabschlag II' and value > -100000;""")


def downgrade():
    pass



"""Add missing foreignkey constraints.

Revision ID: 2fd485c59f80
Revises: db99756d70b
Create Date: 2015-12-18 14:40:07.640937

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '2fd485c59f80'
down_revision = 'db99756d70b'


def upgrade():
    op.execute("""DELETE FROM booking WHERE member_id = 119;""")
    op.execute("""UPDATE booking SET kind_id = 3 WHERE kind_id = 9;""")
    op.execute("""UPDATE booking SET kind_id = 4 WHERE kind_id = 10;""")
    op.create_foreign_key(
        None, 'booking', 'booking', ['splitted_from_id'], ['id'])
    op.create_foreign_key(None, 'booking', 'member', ['member_id'], ['id'])
    op.create_foreign_key(None, 'booking', 'bookingkind', ['kind_id'], ['id'])
    op.create_foreign_key(
        None, 'bookingkind', 'bankingaccount', ['banking_account_id'], ['id'])
    op.create_foreign_key(
        None, 'electricmeter', 'member', ['discount_to_id'], ['id'])
    op.create_foreign_key(None, 'energyvalue', 'member', ['member_id'], ['id'])
    op.create_foreign_key(None, 'formletter', 'user', ['user_id'], ['id'])
    op.create_foreign_key(None, 'message', 'user', ['user_id'], ['id'])


def downgrade():
    op.drop_constraint(None, 'message', type_='foreignkey')
    op.drop_constraint(None, 'formletter', type_='foreignkey')
    op.drop_constraint(None, 'energyvalue', type_='foreignkey')
    op.drop_constraint(None, 'electricmeter', type_='foreignkey')
    op.drop_constraint(None, 'bookingkind', type_='foreignkey')
    op.drop_constraint(None, 'booking', type_='foreignkey')
    op.drop_constraint(None, 'booking', type_='foreignkey')
    op.drop_constraint(None, 'booking', type_='foreignkey')



"""Add direct debit marker to members.

Revision ID: dea40636fb3
Revises: 1c44d30192ea
Create Date: 2015-01-29 09:55:08.515004

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'dea40636fb3'
down_revision = '1c44d30192ea'


def upgrade():
    op.add_column(
        'member', sa.Column('direct_debit', sa.Boolean(), nullable=True))
    op.execute("""UPDATE member set direct_debit = true;""")
    op.execute(
        """UPDATE member set direct_debit = false WHERE id in (
            73, 86, 87, 104, 105, 113, 118, 103, 122, 123, 132, 133, 142, 144,
            145, 147, 150, 151, 156, 163, 164, 177, 179);""")


def downgrade():
    op.drop_column('member', 'direct_debit')



"""Update user model.

Revision ID: 4369f16d7e67
Revises: 43245c2d427e
Create Date: 2015-11-16 10:43:23.539911

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '4369f16d7e67'
down_revision = '43245c2d427e'


def upgrade():
    op.add_column(
        'message',
        sa.Column('user_id', sa.Integer(), nullable=False, server_default='2'))


def downgrade():
    op.drop_column('message', 'user_id')



"""Update member chaos

Revision ID: 5fe4feac157
Revises: 17b7b9cfd7e8
Create Date: 2015-04-01 12:41:15.669249

"""

# revision identifiers, used by Alembic.
from alembic import op

revision = '5fe4feac157'
down_revision = '17b7b9cfd7e8'


def upgrade():
    op.execute("""UPDATE allotment SET member_id = 186 WHERE id = 118;""")
    op.execute("""UPDATE allotment SET member_id = 109 WHERE id = 128;""")
    op.execute("""UPDATE booking SET member_id = 186 WHERE id = 251;""")
    op.execute("""UPDATE booking SET member_id = 186 WHERE id = 250;""")
    op.execute("""UPDATE booking SET member_id = 186 WHERE id = 102;""")
    op.execute("""UPDATE member SET iban = null, bic = null,
                  direct_debit_date = null, direct_debit = false
                  WHERE id = 109;""")
    op.execute("""UPDATE energyvalue SET member_id = 186 WHERE id = 77;""")
    op.execute("""UPDATE energyvalue SET member_id = 186 WHERE id = 78;""")
    op.execute("""UPDATE energyvalue SET member_id = 186 WHERE id = 79;""")
    op.execute("""UPDATE energyvalue SET member_id = 186 WHERE id = 80;""")
    op.execute("""UPDATE energyvalue SET member_id = 109 WHERE id = 101;""")
    op.execute("""UPDATE energyvalue SET member_id = 109 WHERE id = 102;""")
    op.execute("""UPDATE member SET iban = 'DE68800636480801753000',
                  bic = 'GENODEF1NMB', direct_debit_date = '2015-03-30',
                  direct_debit = true WHERE lastname = 'Maxis';""")


def downgrade():
    pass



"""Add protocols

Revision ID: 1850226c55dc
Revises: 48019c236591
Create Date: 2014-12-08 18:52:13.291970

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '1850226c55dc'
down_revision = '48019c236591'


def upgrade():
    op.create_table(
        'protocol',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('title', sa.String(length=254), nullable=True),
        sa.Column('location', sa.String(length=254), nullable=True),
        sa.Column('attendees', sa.String(length=254), nullable=True),
        sa.Column('day', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('protocol')



# coding:utf8
from __future__ import unicode_literals
from babel.numbers import format_currency


def format_kwh(value):
    return '{} kWh'.format(value if value is not None else '--')


def format_eur(value, full=False):
    if value is None:
        return '---.-- €'
    value = value / 10000.0
    format = '#,##0.00 ¤'
    if full:
        format = '#,##0.0000 ¤'
    return format_currency(
        value, 'EUR', format, currency_digits=False, locale='de_DE')



"""Add member to bookings.

Revision ID: eb7a1761a6d
Revises: 28e32b8a26f4
Create Date: 2015-03-10 16:38:06.335776

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = 'eb7a1761a6d'
down_revision = '28e32b8a26f4'


def upgrade():
    op.add_column(
        'booking', sa.Column('member_id', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('booking', 'member_id')



"""Add waiting status to form letters.

Revision ID: 24b5e5215f2d
Revises: 179b760ddcda
Create Date: 2015-01-04 14:01:42.561563

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '24b5e5215f2d'
down_revision = '179b760ddcda'

OLD_STATUS = (u'In Bearbeitung', u'Fehler', u'Fertig gestellt')
NEW_STATUS = (u'Wartet',) + OLD_STATUS


def upgrade():
    op.drop_column('formletter', 'status')
    sa.Enum(name='form_letter_status').drop(
        op.get_bind(), checkfirst=False)
    sa.Enum(*NEW_STATUS, name='form_letter_status').create(
        op.get_bind(), checkfirst=False)
    op.add_column(
        'formletter',
        sa.Column('status', sa.Enum(*NEW_STATUS, name='form_letter_status'),
                  nullable=False))


def downgrade():
    op.drop_column('formletter', 'status')
    sa.Enum(name='form_letter_status').drop(
        op.get_bind(), checkfirst=False)
    sa.Enum(*OLD_STATUS, name='form_letter_status').create(
        op.get_bind(), checkfirst=False)
    op.add_column(
        'formletter',
        sa.Column('status', sa.Enum(*OLD_STATUS, name='form_letter_status'),
                  nullable=False))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .model import Object
from sqlalchemy import (
    Column, Integer, Text, String, ForeignKey, LargeBinary, DateTime)
from .depot import MAX_FILE_SIZE, get_current_user
import datetime
import sqlalchemy.orm


class Bulletin(Object):
    """Ein Aushang."""

    subject = Column(String)
    content = Column(Text)
    mimetype = Column(String(30), default=u'')
    size = Column(String(20), default=u'')
    data = Column(LargeBinary(MAX_FILE_SIZE))
    date = Column(DateTime, nullable=False, default=datetime.datetime.now)
    user_id = Column(Integer, ForeignKey('user.id'), nullable=False,
                     default=get_current_user)
    user = sqlalchemy.orm.relation(
        'User', uselist=False, backref='bulletins')



"""Add comment and estimate value to electic meter

Revision ID: 7136a99f680d
Revises: 95c83e9e7d18
Create Date: 2016-06-26 15:08:42.149634

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '7136a99f680d'
down_revision = '95c83e9e7d18'


def upgrade():
    op.add_column('electricmeter',
                  sa.Column('comment', sa.String(), nullable=True))
    op.add_column(
        'energyvalue',
        sa.Column('estimated_value', sa.Boolean(), nullable=True,
                  server_default='false'))
    op.execute("""UPDATE electricmeter SET comment = 'Wasserpumpe' WHERE id
               in (40, 49, 62, 67, 89)""")
    op.execute("UPDATE electricmeter SET comment = 'Satanlage' WHERE id = 17")
    op.execute("""UPDATE electricmeter SET number = '424690' WHERE id = 62""")
    op.execute("UPDATE electricmeter SET number = '31458149' WHERE id = 61")
    for id in (11, 12, 13, 19, 22, 23, 28, 35, 114, 115, 54, 57, 60, 63, 66,
               72, ):
        op.execute("""UPDATE energyvalue SET estimated_value = true WHERE
                   electric_meter_id = %s and year = 2015""" % id)


def downgrade():
    op.drop_column('energyvalue', 'estimated_value')
    op.drop_column('electricmeter', 'comment')



"""Make haus_nr non unique.

Revision ID: 29f31fb4aa
Revises: 4d42c583ac73
Create Date: 2014-10-27 14:52:05.655804

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '29f31fb4aa'
down_revision = '4d42c583ac73'


def upgrade():
    op.drop_constraint(u'member_haus_nr_key', 'member')


def downgrade():
    op.create_unique_constraint(u'member_haus_nr_key', 'member', ['haus_nr'])



# encoding=utf-8
"""Add position and signature to users.

Revision ID: 2fd64f1e524c
Revises: 24b5e5215f2d
Create Date: 2015-01-04 15:20:53.561304

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '2fd64f1e524c'
down_revision = '24b5e5215f2d'


def upgrade():
    op.add_column(
        'user',
        sa.Column('position', sa.String(length=50), nullable=True))
    op.add_column(
        'user',
        sa.Column('signature', sa.LargeBinary(length=10485760), nullable=True))
    op.execute(u"""UPDATE public.user SET
                  position = 'Schatzmeister',
                  anrede = 'Frau',
                  vorname = 'Susann',
                  nachname = 'Tillack',
                  ort = 'Berlin'
                  WHERE username = 'st';""")
    op.execute(u"""UPDATE public.user SET
                  position = 'Vorsitzender',
                  anrede = 'Herr',
                  vorname = 'Gerd',
                  nachname = 'Mittag',
                  ort = 'Bad Dürrenberg'
                  WHERE username = 'gm';""")
    op.execute(u"""UPDATE public.user SET
                  position = 'Stellvertreter',
                  anrede = 'Frau',
                  vorname = 'Annette',
                  nachname = 'Rösler',
                  ort = 'Bad Dürrenberg'
                  WHERE username = 'ar';""")
    op.execute(u"""UPDATE public.user SET
                  position = 'Schriftführer',
                  anrede = 'Herr',
                  vorname = 'Sebastian',
                  nachname = 'Wehrmann',
                  ort = 'Halle (Saale)'
                  WHERE username = 'sw';""")


def downgrade():
    op.drop_column('user', 'signature')
    op.drop_column('user', 'position')



# encoding=utf8
from .base import to_string, format_eur, format_kwh, boolean, get_selected_year
from .letter import send_mail
from pyramid.view import view_config
from sw.allotmentclub import ElectricMeter, EnergyValue, Allotment, Member
from sw.allotmentclub import EnergyPrice, User
import datetime
import json
import sw.allotmentclub.browser.member
import transaction


def get_discount_to(electric_meter_id, request=None):
    meter = ElectricMeter.get(electric_meter_id)
    allotment = meter.allotment
    if meter.discount_to:
        member = meter.discount_to
    else:
        member = meter.allotment.member
    return '{} ({}, {})'.format(allotment.number, member.lastname,
                                member.firstname)


class Query(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Kraftstrom': boolean,
        'Abgeklemmt': boolean,
        'Bungalow': get_discount_to,
    }
    data_class = {
        'Bungalow': 'expand'
    }
    data_hide = {
        'Kraftstrom': 'phone,tablet',
        'Abgeklemmt': 'phone,tablet',
        'Zählernummer': 'phone',
        'Anmerkung': 'phone',
    }

    def select(self):
        return (
            self.db.query(
                ElectricMeter.id.label('#'),
                ElectricMeter.number.label('Zählernummer'),
                ElectricMeter.id.label('Bungalow'),
                ElectricMeter.electric_power.label('Kraftstrom'),
                ElectricMeter.disconnected.label('Abgeklemmt'),
                ElectricMeter.comment.label('Anmerkung'),
            )
            .select_from(ElectricMeter)
            .outerjoin(Allotment)
            .outerjoin(Member))


@view_config(route_name='electricity_list', renderer='json',
             permission='view')
class ElectricityListView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    default_order_by = 'electricmeter.id'
    available_actions = [
        dict(url='energy_value_list', btn_class='btn-success',
             icon='fa fa-list', title='Energieabrechnungen'),
        dict(url='advance_pay_value_list', btn_class='btn-success',
             icon='fa fa-list', title='Abschläge'),
        dict(url='energy_meter_export', btn_class='btn-success',
             icon='fa fa-cloud-download', title='Exportieren'),
        dict(url='energy_meter_import', btn_class='btn-success',
             icon='fa fa-cloud-upload', title='Importieren')]


def get_to_pay(energy_value_id, request=None):
    value = EnergyValue.get(energy_value_id)
    member = value.member
    allotments = '/'.join(str(a.number) for a in member.allotments)
    if not allotments:
        allotments = 'n/a'
    return '{} ({}, {})'.format(allotments, member.lastname, member.firstname)


class GlobalEnergyValueQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Verbrauch': format_kwh,
        'Kosten': format_eur,
        'Gebühr': format_eur,
        'Gesamt': format_eur,
        'Offen': format_eur,
        'Abschlag': format_eur,
        'Bungalow': get_to_pay,
    }
    css_classes = {
        'Kosten': 'right',
        'Gebühr': 'right',
        'Verbrauch': 'right',
        'Gesamt': 'right',
        'Offen': 'right',
        'Abschlag': 'right',
    }
    data_class = {
        'Bungalow': 'expand'
    }
    data_hide = {
        'Zähler': 'phone',
        'Verbrauch': 'phone,tablet',
        'Kosten': 'phone,tablet',
        'Gebühr': 'phone,tablet',
        'Gesamt': 'phone',
        'Abschlag': 'phone',
    }

    def select(self):
        return (
            self.db.query(
                EnergyValue.id.label('#'),
                EnergyValue.id.label('Bungalow'),
                ElectricMeter.number.label('Zähler'),
                EnergyValue.usage.label('Verbrauch'),
                EnergyValue.price.label('Kosten'),
                EnergyValue.fee.label('Gebühr'),
                EnergyValue.whole_price.label('Gesamt'),
                EnergyValue.to_pay.label('Offen'),
                EnergyValue.advance_pay.label('Abschlag'),
            )
            .select_from(EnergyValue)
            .outerjoin(ElectricMeter)
            .group_by(EnergyValue.id, ElectricMeter.id)
            .filter(EnergyValue.year == get_selected_year()))


@view_config(route_name='global_energy_value_list', renderer='json',
             permission='view')
class GlobalEnergyValueListView(sw.allotmentclub.browser.base.TableView):

    query_class = GlobalEnergyValueQuery
    year_selection = True
    start_year = 2013
    available_actions = [
        dict(url='calculate_energy_values', btn_class='btn-danger',
             icon='fa fa-refresh', title='Neuberechnen'),
    ]


def rboolean(value, *args, **kw):
    return boolean(not value, *args, **kw)


class EnergyValueQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Zählerstand': format_kwh,
        'Abgelesen': rboolean,
        'Verbrauch': format_kwh,
        'Kosten': format_eur,
        'Gebühr': format_eur,
        'Gesamt': format_eur,
        'Offen': format_eur,
        'Abgerechnet': boolean,
    }
    data_class = {
        'Jahr': 'expand'
    }
    css_classes = {
        'Kosten': 'right',
        'Gebühr': 'right',
        'Verbrauch': 'right',
        'Offen': 'right',
    }
    data_hide = {
        'Zählerstand': 'phone,tablet',
        'Abgelesen': 'phone, tablet',
        'Verbrauch': 'phone',
        'Kosten': 'phone,tablet',
        'Gebühr': 'phone',
        'Rechnung an': 'phone,tablet',
        'Abgerechnet': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                EnergyValue.id.label('#'),
                EnergyValue.year.label('Jahr'),
                EnergyValue.value.label('Zählerstand'),
                EnergyValue.estimated_value.label('Abgelesen'),
                EnergyValue.usage.label('Verbrauch'),
                EnergyValue.price.label('Kosten'),
                EnergyValue.fee.label('Gebühr'),
                EnergyValue.whole_price.label('Gesamt'),
                EnergyValue.to_pay.label('Offen'),
                (to_string(Member.lastname).concat(', ')
                 .concat(to_string(Member.firstname))
                 .label('Rechnung an')),
                EnergyValue.discounted.label('Abgerechnet'),
            )
            .select_from(EnergyValue)
            .outerjoin(Member)
            .filter(EnergyValue.electric_meter == self.context))


@view_config(route_name='energy_value_list', renderer='json',
             permission='view')
class EnergyValueListView(sw.allotmentclub.browser.base.TableView):

    query_class = EnergyValueQuery
    default_order_by = 'year'


@view_config(route_name='calculate_energy_values', renderer='json',
             permission='view')
class CalculateEnergyValues(sw.allotmentclub.browser.base.View):

    def update_price(self):
        org_id = self.request.user.organization_id
        current_year = get_selected_year()
        price = (EnergyPrice.query()
                 .filter(EnergyValue.organization_id == org_id)
                 .filter(EnergyPrice.year == current_year).one())
        price.usage_hauptzaehler = price.value - (
            EnergyPrice.query()
            .filter(EnergyValue.organization_id == org_id)
            .filter(EnergyPrice.year == current_year-1).one().value)
        price.usage_members = sum(
            e.usage for e in (EnergyValue.query()
                              .filter(EnergyValue.year == current_year)
                              .filter(EnergyValue.organization_id == org_id)))
        price.leakage_current = price.usage_hauptzaehler - price.usage_members
        price.price = price.bill / price.usage_hauptzaehler
        self.phases = sum(
            3 if e.electric_power else 1
            for e in (
                ElectricMeter.query()
                .filter(ElectricMeter.disconnected.is_(False))
                .filter(ElectricMeter.organization_id == org_id)))
        price.normal_fee = price.leakage_current * price.price / self.phases
        price.power_fee = price.normal_fee * 3

    def send_report(self):
        org_id = self.request.user.organization_id
        current_year = get_selected_year()
        price = (EnergyPrice.query()
                 .filter(EnergyValue.organization_id == org_id)
                 .filter(EnergyPrice.year == current_year).one())

        data = json.loads(json.dumps(price))
        data['phases'] = self.phases
        data['price'] = format_eur(data['price'])
        data['bill'] = format_eur(data['bill'])
        data['normal_fee'] = format_eur(data['normal_fee'])
        data['power_fee'] = format_eur(data['power_fee'])

        subject = "Bericht Energieabrechnung {year}".format(**data)

        body = """
        <p>Sehr geehrter Vorstand,</p>

        <p>anbei die Kennzahlen zur diesjährigen Energieabrechnung:</p>

        <table>
        <tr><td>Verbrauch Hauptzähler: </td><td>{usage_hauptzaehler} kWh</td></tr>
        <tr><td>Verbrauch der Mitglieder (inkl. SAT-Anlage und
        Vereinsgebäude): </td><td>{usage_members} kWh</td></tr>
        <tr><td colspan=2></td></tr>

        <tr><td>Verluststrom: </td><td>{usage_hauptzaehler} kWh - {usage_members} kWh = {leakage_current} kWh</td></tr>
        <tr><td colspan=2></td></tr>

        <tr><td>Preis pro kWh: </td><td>{bill} / {usage_hauptzaehler} kWh = {price}/kWh</td></tr>
        <tr><td colspan=2></td></tr>

        <tr><td>Grundgebühr 1 Phase ({phases} insgesamt): </td><td>{leakage_current} kWh * {price}/kWh / {phases} = {normal_fee}</td></tr>
        <tr><td>Grundgebühr einfacher Zähler (1 Phase): </td><td>{normal_fee}</td></tr>
        <tr><td>Grundgebühr Kraftstromzähler (3 Phase): </td><td>{power_fee}</td></tr>
        </table><br /><br />""".format(**data)  # noqa
        send_mail(
            "vorstand@roter-see.de", subject, body, User.by_username('system'))

    def update(self):
        self.update_price()
        values = (EnergyValue.query()
                  .filter(EnergyValue.discounted.is_(None))
                  .filter(EnergyValue.year == get_selected_year())
                  .all())
        for value in values:
            value.update_data()
        self.send_report()


class AdvancePayValueQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Zählerstand': format_kwh,
        'Verbrauch': format_kwh,
        'Abschlag': format_eur,
        'Gesamt': format_eur,
    }
    data_class = {
        'Jahr': 'expand'
    }
    css_classes = {
        'Abschlag': 'right',
        'Verbrauch': 'right',
    }
    data_hide = {
        'Zählerstand': 'phone,tablet',
        'Verbrauch': 'phone',
        'Abschlag': 'phone',
        'Rechnung an': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                EnergyValue.id.label('#'),
                EnergyValue.year.label('Jahr'),
                EnergyValue.value.label('Zählerstand'),
                EnergyValue.usage.label('Verbrauch'),
                EnergyValue.whole_price.label('Gesamt'),
                EnergyValue.advance_pay.label('Abschlag'),
                (to_string(Member.lastname).concat(', ')
                 .concat(to_string(Member.firstname))
                 .label('Rechnung an')),
            )
            .select_from(EnergyValue)
            .outerjoin(Member)
            .filter(EnergyValue.electric_meter == self.context))


@view_config(route_name='advance_pay_value_list', renderer='json',
             permission='view')
class AdvancePayValueListView(sw.allotmentclub.browser.base.TableView):

    query_class = AdvancePayValueQuery
    default_order_by = 'year'


@view_config(route_name='energy_meter_export', renderer='json',
             permission='view')
class EnergyMeterExporterView(sw.allotmentclub.browser.base.XLSXExporterView):

    filename = 'Stromzaehler'
    title = 'Zählerstände'

    def query(self):
        from sqlalchemy.sql.expression import false
        current_year = get_selected_year()
        return (
            self.db.query(
                ElectricMeter.id.label('#'),
                ElectricMeter.id.label('Nachname'),
                ElectricMeter.id.label('Vorname'),
                Allotment.number.label('Bungalow'),
                ElectricMeter.number.label('Zählernummer'),
                ElectricMeter.id.label('Zähler {}'.format(current_year-3)),
                ElectricMeter.id.label('Zähler {}'.format(current_year-2)),
                ElectricMeter.id.label('Zähler {}'.format(current_year-1)),
                ElectricMeter.id.label('Zähler {}'.format(current_year)),
            )
            .select_from(ElectricMeter)
            .join(Allotment)
            .join(Member)
            .order_by(Allotment.number)
            .filter(ElectricMeter.disconnected == false())
            .filter(ElectricMeter.id != 124)
        )

    def _data(self, query):
        data = super(EnergyMeterExporterView, self)._data(query)
        for item in data:
            item = list(item)
            meter = ElectricMeter.get(item[1])
            item[1] = meter.allotment.member.lastname
            item[2] = meter.comment or meter.allotment.member.firstname
            year = get_selected_year()
            value_sec3last_year = meter.get_value(year-3)
            value_sec2last_year = meter.get_value(year-2)
            value_last_year = meter.get_value(year-1)
            value_current_year = meter.get_value(year)
            item[-4] = self._get_value(value_sec3last_year)
            item[-3] = self._get_value(value_sec2last_year)
            item[-2] = self._get_value(value_last_year)
            item[-1] = self._get_value(value_current_year)
            yield item
        item = [0, 'Hauptzähler', 'kWh', 0, 48328153, '', '', '', '']
        for i in range(-1, -5, -1):
            price = (EnergyPrice.query()
                     .filter(EnergyPrice.year == year+i+1).first())
            if price:
                item[i] = price.value
        yield item
        item = [0, 'Endabrechnung', 'microcent', 0, 48328153, '', '', '', '']
        for i in range(-1, -5, -1):
            price = (EnergyPrice.query()
                     .filter(EnergyPrice.year == year+i+1).first())
            if price:
                item[i] = price.bill
        yield item

    def after_export(self):
        red_format = self.workbook.add_format(
            {'bold': True, 'font_color': 'red'})
        irow = self.worksheet.dim_colmax + 1
        lrow = chr(ord('A') + irow)
        self.worksheet.write(1, irow, 'Verbrauch', self.bold_style)
        for i in range(self.worksheet.dim_rowmax - 2):
            i += 3
            cellindex = '{}{}'.format(lrow, i)
            formula = '=IF({r1}{i},{r1}{i}-{r2}{i},0)'.format(
                i=i, r1=chr(ord('A') + irow-1), r2=chr(ord('A') + irow-2))
            self.worksheet.write_formula(cellindex, formula)
            self.worksheet.conditional_format(
                cellindex, {'type': 'cell', 'criteria': '<', 'value': 0,
                            'format': red_format})

        irow += 1
        lrow = chr(ord('A') + irow)
        self.worksheet.write(1, irow, 'Zähler gesehen?', self.bold_style)
        for i in range(self.worksheet.dim_rowmax - 2):
            i += 3
            cellindex = '{}{}'.format(lrow, i)
            self.worksheet.write_boolean(cellindex, False)
            self.worksheet.data_validation(
                cellindex, {'validate': 'list', 'source': ['WAHR', 'FALSCH']})

    def _get_value(self, value_item):
        if not value_item:
            return ''
        if value_item.estimated_value:
            return (value_item.value, {'color': '#8D0009'})
        return value_item.value


@view_config(route_name='energy_meter_import',
             permission='view',
             renderer='json')
class EnergyMeterImporterView(sw.allotmentclub.browser.base.XLSXImporterView):

    cell_index = 8

    def __call__(self):
        year = datetime.datetime.now().year
        # Sonderbehandlung Satellitenanlage
        meter = ElectricMeter.get(124)
        if meter and not meter.get_value(year):
            last = meter.get_last_value()
            energy_value = EnergyValue.create(
                electric_meter=meter, year=year, value=last.value + 552)
            energy_value.update_member()
            energy_value.update_usage()  # verbrauch berechnen
            if energy_value not in meter.energy_values:
                meter.energy_values.append(energy_value)
        return super(EnergyMeterImporterView, self).__call__()

    def add_data(self, line):
        year = datetime.datetime.now().year
        if line[1].value == 'Hauptzähler':
            EnergyPrice.find_or_create(year=year).value = line[
                self.cell_index].value
            transaction.savepoint()
            return
        if line[1].value == 'Endabrechnung':
            EnergyPrice.find_or_create(year=year).bill = line[
                self.cell_index].value
            transaction.savepoint()
            return
        if line[0].value is None and line[1].value is None:
            return
        meter = ElectricMeter.by_number(line[4].value)
        if not line[self.cell_index].value:
            value = int(line[self.cell_index-1].value)
            estimated = True
        else:
            value = int(line[self.cell_index].value)
            estimated = not line[self.cell_index+2].value
        energy_value = EnergyValue.create(
            electric_meter=meter,
            year=year,
            value=value,
            estimated_value=estimated)
        if energy_value not in meter.energy_values:
            meter.energy_values.append(energy_value)
        energy_value.update_member()
        energy_value.update_usage()  # verbrauch berechnen
        transaction.savepoint()


class EnergyPriceQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Stand Hauptzähler': format_kwh,
        'Verbrauch Hauptzähler': format_kwh,
        'Verbrauch Mitglieder': format_kwh,
        'Verluststrom': format_kwh,
        'Preis kWh': format_eur,
        'Grundgebühr': format_eur,
        'Grundgebühr Starkstrom': format_eur,
        'Endabrechnung Energieanbieter': format_eur,
    }
    css_classes = {
        'Stand Hauptzähler': 'right',
        'Verbrauch Hauptzähler': 'right',
        'Verbrauch Mitglieder': 'right',
        'Verluststrom': 'right',
        'Preis kWh': 'right',
        'Grundgebühr': 'right',
        'Grundgebühr Starkstrom': 'right',
        'Endabrechnung Energieanbieter': 'right',
    }
    data_class = {
        'Preis kWh': 'expand'
    }
    data_hide = {
        'Grundgebühr': 'phone',
        'Grundgebühr Starkstrom': 'phone,tablet',
        'Stand Hauptzähler': 'phone,tablet',
        'Verbrauch Hauptzähler': 'phone,tablet',
        'Verluststrom': 'phone',
    }

    def select(self):
        return (
            self.db.query(
                EnergyPrice.id.label('#'),
                EnergyPrice.price.label('Preis kWh'),
                EnergyPrice.normal_fee.label('Grundgebühr'),
                EnergyPrice.power_fee.label('Grundgebühr Starkstrom'),
                EnergyPrice.value.label('Stand Hauptzähler'),
                EnergyPrice.usage_hauptzaehler.label('Verbrauch Hauptzähler'),
                EnergyPrice.usage_members.label('Verbrauch Mitglieder'),
                EnergyPrice.leakage_current.label('Verluststrom'),
                EnergyPrice.bill.label('Endabrechnung Energieanbieter'),
            )
            .select_from(EnergyPrice)
            .filter(EnergyPrice.year == get_selected_year()))


@view_config(route_name='energy_price', renderer='json', permission='view')
class EnergyPriceListView(sw.allotmentclub.browser.base.TableView):

    query_class = EnergyPriceQuery
    year_selection = True
    start_year = 2014



"""Add protocol attachments and commitments.

Revision ID: 51143cacbf85
Revises: 571eb56a2b04
Create Date: 2014-12-17 11:31:54.057025

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '51143cacbf85'
down_revision = '571eb56a2b04'


def upgrade():
    op.create_table(
        'protocolcommitment',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('protocol_id', sa.Integer(), nullable=False),
        sa.Column('who', sa.String(length=254), nullable=True),
        sa.Column('what', sa.String(length=254), nullable=True),
        sa.Column('when', sa.String(length=254), nullable=True),
        sa.ForeignKeyConstraint(['protocol_id'], [u'protocol.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'protocolattachment',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('protocol_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=100), nullable=True),
        sa.Column('mimetype', sa.String(length=30), nullable=True),
        sa.Column('size', sa.String(length=20), nullable=True),
        sa.Column('data', sa.LargeBinary(length=10485760), nullable=True),
        sa.ForeignKeyConstraint(['protocol_id'], [u'protocol.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('protocolattachment')
    op.drop_table('protocolcommitment')



# -*- coding: utf-8 -*-
from sw.allotmentclub.model import ENGINE_NAME
import risclog.sqlalchemy.db


class Application(object):

    settings = None
    testing = False

    @classmethod
    def from_filename(cls, filename):
        """Creates a portal from a paste.ini file."""
        # We have to do the import here because the converter cannot depend
        # on pyramid:
        import pyramid.paster
        pyramid.paster.setup_logging(filename)
        settings = pyramid.paster.get_appsettings(filename)
        portal = cls(**settings)
        portal.configure()
        portal.setup_runtime()
        return portal

    def __init__(self, testing=False, **settings):
        self.settings = settings
        self.settings['testing'] = testing
        self.testing = testing

    def __call__(self, **settings):
        self.settings.update(settings)
        self.configure()
        self.setup_runtime()
        # can't go into setup_runtime, since initialize_db uses that, too
        self.check_db_revision()

    def setup_runtime(self):
        if not self.testing:
            db = risclog.sqlalchemy.db.get_database()
            db.register_engine(self.settings['sqlalchemy.url'],
                               engine_args=dict(pool_pre_ping=True),
                               name=ENGINE_NAME,
                               alembic_location='sw.allotmentclub:alembic')

    def check_db_revision(self):
        db = risclog.sqlalchemy.db.get_database(self.testing)
        try:
            db.assert_database_revision_is_current(ENGINE_NAME)
        except ValueError:
            # Bail out when database revision does not match.
            raise SystemExit(
                "Database revision does not match current head. "
                "Please run alembic to get migrated to the latest revision.")

    def configure(self):
        pass



"""Add leaving year to members.

Revision ID: 55a653c1d5a0
Revises: 30de960848fa
Create Date: 2015-11-06 14:11:22.301052

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '55a653c1d5a0'
down_revision = u'30de960848fa'


def upgrade():
    op.add_column(
        'member',
        sa.Column('leaving_year', sa.Integer(), nullable=True))
    op.execute("""UPDATE member SET leaving_year = 2014 WHERE id = 190;""")
    op.execute("""UPDATE member SET leaving_year = 2015 WHERE id in (
    142, 146, 128, 132, 86, 176);""")


def downgrade():
    op.drop_column('member', 'leaving_year')



# encoding: utf-8
from sw.allotmentclub import DashboardData
from pyramid.view import view_config
import sw.allotmentclub.browser.base
import sw.allotmentclub.browser.login
import time


@view_config(route_name='dashboard', renderer='json',
             permission='view')
class DashboardListView(sw.allotmentclub.browser.base.View,
                        sw.allotmentclub.browser.login.NetatmoMixin):

    temperature_data = (
        ('luefter_aussenluft', 'Aussenluft', '#E24913', '{} °C'),
        ('luefter_zuluft', 'Zuluft', '#6595b4', '{} °C'),
        ('luefter_fortluft', 'Fortluft', '#FF9F01', '{} °C'),
        ('luefter_abluft', 'Abluft', '#7e9d3a', '{} °C'),
        ('luefter_abluft_feuchte', 'Abluft Feuchte', '#7e9d3a', '{} %'),
        ('luefter_percent', 'Lüfterstufe', '#BD362F', '{} %'),
    )

    plot_data = ('luefter_aussenluft', 'luefter_zuluft', 'luefter_fortluft',
                 'luefter_abluft')

    def update(self):
        result = dict(temperatures=[], plots=[])
        for i, d in enumerate(
                DashboardData.query()
                .order_by(DashboardData.date.desc()).limit(576)):
            unixtime = time.mktime(d.date.timetuple()) * 1000
            if i == 0:
                result['date'] = d.date.strftime('am %d.%m.%Y um %H:%M Uhr')
                for k, title, color, fmt in self.temperature_data:
                    result['temperatures'].append(dict(
                        title=title,
                        color=color,
                        current_value=getattr(d, k),
                        tendency='',
                        current_value_str=fmt.format(getattr(d, k))))
                    result['plots'].append([])
            if i == 1:
                for j, (k, title, color, fmt) in enumerate(
                        self.temperature_data):
                    tendency = 'right'
                    if (result['temperatures'][j]['current_value'] <
                            getattr(d, k)):
                        tendency = 'down'
                    elif (result['temperatures'][j]['current_value'] >
                            getattr(d, k)):
                        tendency = 'up'
                    result['temperatures'][j]['tendency'] = tendency

            for j, k in enumerate(self.plot_data):
                result['plots'][j].insert(0, [unixtime, getattr(d, k)])
        result['temp'] = self.get_temp_data(dest='wachtelberg')
        self.result = dict(data=result, status='success')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from io import BytesIO
from pyramid.response import FileIter
from .base import date_time, get_selected_year, datetime_now, format_size
from .base import format_mimetype
from ..log import user_data_log, log_with_user
from pyramid.view import view_config
import datetime
from sw.allotmentclub import (
    Protocol, ProtocolDetail, ProtocolAttachment, ProtocolCommitment)
import collections
import dateutil.parser
import sw.allotmentclub.browser.base
import markdown


def int2roman(n):
    roman_numerals = []
    zahlen = {1000: "M",
              900: "CM",
              500: "D",
              400: "CD",
              100: "C",
              90: "XC",
              50: "L",
              40: "XL",
              10: "X",
              9: "IX",
              5: "V",
              4: "IV",
              1: "I"}

    if n > 0 and n < 4000 and type(n) == int:
        for zahl in sorted(zahlen.keys(), reverse=True):
            roman_numerals.append(n // zahl * zahlen[zahl])
            n = n % zahl

        return "".join(roman_numerals)


class Query(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Datum': date_time
    }

    data_class = {
        'Datum': 'expand'
    }
    data_hide = {
        'Titel': 'phone,tablet',
        'Teilnehmer': 'phone,tablet',
        'Ort': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                Protocol.id.label('#'),
                Protocol.day.label('Datum'),
                Protocol.title.label('Titel'),
                Protocol.attendees.label('Teilnehmer'),
                Protocol.location.label('Ort'))
            .select_from(Protocol)
            .filter(Protocol.accounting_year == get_selected_year()))


@view_config(route_name='protocols', renderer='json',
             permission='view')
class ProtocolListView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Protokolle."""

    query_class = Query
    default_order_by = 'day'
    year_selection = True
    start_year = 2014
    available_actions = [
        dict(url='protocol_add', btn_class='btn-success', icon='fa fa-plus',
             title='Neu'),
        dict(url='protocol_edit', btn_class='btn-success', icon='fa fa-pencil',
             title='Bearbeiten'),
        dict(url='protocol_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen'),
        dict(url='protocol_detail', btn_class='btn-success', icon='fa fa-list',
             title='Details'),
        dict(url='protocol_attachment', btn_class='btn-success',
             icon='fa fa-list', title='Anlagen'),
        dict(url='protocol_commitment', btn_class='btn-success',
             icon='fa fa-list', title='Absprachen'),
        dict(url='protocol_print', btn_class='btn-success',
             icon='fa fa-print', title='Herunterladen')]


@view_config(route_name='protocol_edit', renderer='json',
             permission='view')
class ProtocolEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Protokoll bearbeiten'

    @property
    def load_options(self):
        return {
            'day': {
                'label': 'Tag',
                'css_class': 'datetimepicker',
            },
            'title': {
                'label': 'Titel',
            },
            'attendees': {
                'label': 'Teilnehmer',
            },
            'location': {
                'label': 'Ort',
            },

        }

    @property
    def load_data(self):
        fields = [
            ('day', date_time(self.context.day)),
            ('title', self.context.title),
            ('attendees', self.context.attendees),
            ('location', self.context.location),
        ]
        return collections.OrderedDict(fields)

    def update(self):
        super(ProtocolEditView, self).update()
        if self.context.day:
            if isinstance(self.context.day, datetime.datetime):
                self.context.accounting_year = self.context.day.year
            else:
                self.context.accounting_year = (
                    dateutil.parser.parse(self.context.day).year)


@view_config(route_name='protocol_add', renderer='json',
             permission='view')
class ProtocolAddView(ProtocolEditView):

    title = 'Protokoll hinzufügen'

    def __init__(self, context, request):
        context = Protocol.create(
            day=datetime_now(),
            accounting_year=get_selected_year())
        context.commit()
        super(ProtocolAddView, self).__init__(context, request)
        log_with_user(user_data_log.info, self.request.user,
                      'Protokoll %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'protocol_edit'


@view_config(route_name='protocol_delete', renderer='json',
             permission='view')
class ProtocolDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = Protocol


def format_markdown(value, request=None):
    if not value:
        return
    return markdown.markdown(value)


class DetailQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'TOP': format_markdown
    }
    data_class = {
        'Dauer': 'expand'
    }
    data_hide = {
        'TOP': 'phone,tablet',
        'Verantwortlich': 'phone',
    }

    def select(self):
        return (
            self.db.query(
                ProtocolDetail.id.label('#'),
                ProtocolDetail.duration.label('Dauer'),
                ProtocolDetail.message.label('TOP'),
                ProtocolDetail.responsible.label('Verantwortlich'))
            .select_from(ProtocolDetail)
            .filter_by(protocol=self.context))


@view_config(route_name='protocol_detail', renderer='json',
             permission='view')
class ProtocolDetailsView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Protokolle."""

    query_class = DetailQuery
    default_order_by = 'id'
    available_actions = [
        dict(url='protocol_detail_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='protocol_detail_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='protocol_detail_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen')]


@view_config(route_name='protocol_detail_edit', renderer='json',
             permission='view')
class ProtocolDetailEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Protokoll-Detail bearbeiten'

    @property
    def load_options(self):
        return {
            'duration': {
                'label': 'Dauer',
            },
            'message': {
                'label': 'TOP',
                'template': 'form_markdown',
            },
            'responsible': {
                'label': 'Verantwortlich',
            },
        }

    @property
    def load_data(self):
        fields = [
            ('duration', self.context.duration),
            ('message', self.context.message),
            ('responsible', self.context.responsible),
        ]
        return collections.OrderedDict(fields)

    def get_route(self, item, name):
        route = super(ProtocolDetailEditView, self).get_route(item, name)
        return route.replace('{protocol_id}', str(self.context.protocol.id))


@view_config(route_name='protocol_detail_add', renderer='json',
             permission='view')
class ProtocolDetailAddView(ProtocolDetailEditView):

    title = 'Protokoll-Detail hinzufügen'

    def __init__(self, context, request):
        context = ProtocolDetail.create(protocol=context)
        context.commit()
        super(ProtocolDetailAddView, self).__init__(context, request)
        log_with_user(user_data_log.info, self.request.user,
                      'ProtokollDetail %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'protocol_detail_edit'


@view_config(route_name='protocol_detail_delete', renderer='json',
             permission='view')
class ProtocolDetailDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = ProtocolDetail

    def log(self):
        if self.deleted is not None:
            deleted = self.context.id
            protocol = self.context.protocol.title
            log_with_user(user_data_log.info,
                          self.request.user,
                          'Protokoll-Detail %s aus Protokoll %s gelöscht.',
                          deleted, protocol)


class AttachmentQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Größe': format_size,
        'Dateityp': format_mimetype,
    }
    data_class = {
        'Name': 'expand'
    }
    data_hide = {
        'Dateityp': 'phone,tablet',
        'Größe': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                ProtocolAttachment.id.label('#'),
                ProtocolAttachment.name.label('Name'),
                ProtocolAttachment.mimetype.label('Dateityp'),
                ProtocolAttachment.size.label('Größe'))
            .select_from(ProtocolAttachment)
            .filter_by(protocol=self.context))


@view_config(route_name='protocol_attachment', renderer='json',
             permission='view')
class ProtocolAttachmentsView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Anlagen."""

    query_class = AttachmentQuery
    default_order_by = 'Name'
    available_actions = [
        dict(url='protocol_attachment_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='protocol_attachment_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen'),
        dict(url='protocol_attachment_download', btn_class='btn-success',
             icon='fa fa-download', title='Herunterladen')]


@view_config(route_name='protocol_attachment_add', renderer='json',
             permission='view')
class ProtocolAttachmentAddView(sw.allotmentclub.browser.base.AddView):

    model = ProtocolAttachment

    def parse_file(self):
        file = self.request.params.get('file')
        data = file.file
        data.seek(0)
        data = data.read()
        name = 'Anlage %s' % int2roman(len(self.context.attachments) + 1)
        mimetype = file.type
        size = len(data)
        return name, mimetype, size, data

    def __call__(self):
        if not self.form_submit():
            return {'status': 'success', 'data': {}}
        name, mimetype, size, data = self.parse_file()
        attachment = self.model.create(
            name=name, mimetype=mimetype, size=size, data=data,
            protocol=self.context)
        attachment.commit()
        log_with_user(user_data_log.info,
                      self.request.user,
                      'hat Anlage %s zu Protokoll %s hinzugefügt.',
                      attachment.id, self.context.id)
        return {'status': 'success'}


@view_config(route_name='protocol_attachment_download',
             permission='view')
class ProtocolAttachmentDownloadView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = self.context.mimetype
        response.content_length = int(self.context.size)
        response.content_disposition = (
            'attachment; filename="{}"'.format(self.context.name))
        response.app_iter = FileIter(BytesIO(self.context.data))
        return response


@view_config(route_name='protocol_attachment_delete', renderer='json',
             permission='view')
class ProtocolAttachmentDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = ProtocolAttachment

    def log(self):
        if self.deleted is not None:
            deleted = self.context.id
            protocol = self.context.protocol.title
            log_with_user(user_data_log.info,
                          self.request.user,
                          'Protokoll-Anlage %s aus Protokoll %s gelöscht.',
                          deleted, protocol)


class CommitmentQuery(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Wer?': 'expand'
    }
    data_hide = {
        'Was?': 'phone,tablet',
        'Bis wann?': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                ProtocolCommitment.id.label('#'),
                ProtocolCommitment.who.label('Wer?'),
                ProtocolCommitment.what.label('Was?'),
                ProtocolCommitment.when.label('Bis wann?'))
            .select_from(ProtocolCommitment)
            .filter_by(protocol=self.context))


@view_config(route_name='protocol_commitment', renderer='json',
             permission='view')
class ProtocolCommitmentsView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Protokolle."""

    query_class = CommitmentQuery
    default_order_by = 'id'
    available_actions = [
        dict(url='protocol_commitment_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='protocol_commitment_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='protocol_commitment_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen')]


@view_config(route_name='protocol_commitment_edit', renderer='json',
             permission='view')
class ProtocolCommitmentEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Protokoll-Absprache bearbeiten'

    @property
    def load_options(self):
        return {
            'who': {
                'label': 'Wer?',
            },
            'what': {
                'label': 'Was?',
            },
            'when': {
                'label': 'Bis wann?',
            },
        }

    @property
    def load_data(self):
        fields = [
            ('who', self.context.who),
            ('what', self.context.what),
            ('when', self.context.when),
        ]
        return collections.OrderedDict(fields)

    def get_route(self, item, name):
        route = super(ProtocolCommitmentEditView, self).get_route(item, name)
        return route.replace('{protocol_id}', str(self.context.protocol.id))


@view_config(route_name='protocol_commitment_add', renderer='json',
             permission='view')
class ProtocolCommitmentAddView(ProtocolCommitmentEditView):

    title = 'Protokoll-Absprache hinzufügen'

    def __init__(self, context, request):
        context = ProtocolCommitment.create(protocol=context)
        context.commit()
        super(ProtocolCommitmentAddView, self).__init__(context, request)
        log_with_user(
            user_data_log.info, self.request.user,
            'ProtokollAbsprache %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'protocol_commitment_edit'


@view_config(route_name='protocol_commitment_delete', renderer='json',
             permission='view')
class ProtocolCommitmentDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = ProtocolCommitment

    def log(self):
        if self.deleted is not None:
            deleted = self.context.id
            protocol = self.context.protocol.title
            log_with_user(user_data_log.info,
                          self.request.user,
                          'Protokoll-Absprache %s aus Protokoll %s gelöscht.',
                          deleted, protocol)



"""Add date and user to message attachments

Revision ID: f4b90b599d2a
Revises: 5baeb56d1864
Create Date: 2019-01-30 11:38:21.775451

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'f4b90b599d2a'
down_revision = '5baeb56d1864'


def upgrade():
    op.add_column('attachment',
                  sa.Column('date', sa.DateTime(), nullable=True,
                            server_default=sa.text('now()')))
    op.add_column(
        'attachment', sa.Column('user_id', sa.Integer(), nullable=True))
    op.create_foreign_key(
        'attachment_user_fkey', 'attachment', 'user', ['user_id'], ['id'])
    op.execute("UPDATE attachment SET date = null;")


def downgrade():
    op.drop_constraint(
        'attachment_user_fkey', 'attachment', type_='foreignkey')
    op.drop_column('attachment', 'user_id')
    op.drop_column('attachment', 'date')



"""Model changes to assignment in order to support jsform.

Revision ID: db99756d70b
Revises: 126c6a200c9f
Create Date: 2015-11-20 12:39:19.449012

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = 'db99756d70b'
down_revision = u'126c6a200c9f'


def upgrade():
    op.alter_column(
        'assignmentattendee', 'hours', existing_type=sa.REAL(), nullable=True)
    op.alter_column(
        'assignmentattendee', 'member_id', existing_type=sa.INTEGER(),
        nullable=True)


def downgrade():
    op.alter_column(
        'assignmentattendee', 'member_id', existing_type=sa.INTEGER(),
        nullable=False)
    op.alter_column(
        'assignmentattendee', 'hours', existing_type=sa.REAL(), nullable=False)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .protocol import format_markdown
from io import BytesIO
from pyramid.view import view_config
from PyPDF2 import PdfFileWriter, PdfFileReader
from .mail import append_pdf
import datetime
import img2pdf
import pybars
import sw.allotmentclub.browser.base
import xhtml2pdf.pisa


TEMPLATE = """
<html lang=de>
  <head>
    <meta charset="utf-8" />
    <style>
      @page {
        size: a4 portrait;
        @frame content_frame {
            left: 50pt; width: 512pt; top: 50pt; height: 692pt;
        }
      }
      body { font-size: 10pt; }
    </style>
  </head>
  <body>
    <h1>{{type}}</h1>
    <h3>{{title}}</h3>
    Datum: {{day}}<br />
    Teilnehmer: {{attendees}}<br />
    Ort: {{location}}<br />
    <hr />
    <table>
      <thead>
        <tr>
          <th style="width: 7%; text-align:left">Zeit</th>
          <th style="width: 85%; text-align:left">TOP</th>
          <th style="width: 8%; text-align:right">Verantw.</th>
        </tr>
      </thead>
    </table>
    <hr />
    {{#each details}}
    <table>
      <tbody>
          <tr>
            <td style="width: 7%; vertical-align: top;">{{time}}</td>
            <td style="width: 85%">{{{message}}}</td>
            <td style="width: 8%; vertical-align: top;">{{{responsible}}}</td>
          </tr>
      </tbody>
    </table>
    <hr />
    {{/each}}

    {{#if commitments}}
    <pdf:nextpage />
    <h3>Absprachen</h3>
    <hr />
    <table>
      <thead>
        <tr>
          <th style="width: 7%; text-align:left">Wer?</th>
          <th style="width: 80%; text-align:left">Was?</th>
          <th style="width: 13%; text-align:left">Bis wann?</th>
        </tr>
      </thead>
      <tbody>
        {{#each commitments}}
          <tr>
            <td style="width: 7%; vertical-align: top;">{{who}}</td>
            <td style="width: 80%">{{{what}}}</td>
            <td style="width: 13%; vertical-align: top;">{{when}}</td>
          </tr>
        {{/each}}
      </tbody>
    </table>
    {{/if}}

    {{#if attachments}}
    <pdf:nextpage />
    <h3>Anlagen</h3>
    {{/if}}
  </body>
</html>"""


@view_config(route_name='protocol_print', permission='view')
class ProtocolPrintView(sw.allotmentclub.browser.base.PrintBaseView):

    filename = 'Protokoll'

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_pdf(self):
        # Base data
        data = self.get_json(self.context)
        # Details data
        data['details'] = []
        time = self.context.day
        minutes = 0
        for detail in self.context.details:
            detail = self.get_json(detail)
            time += datetime.timedelta(minutes=minutes)
            minutes = detail['duration']
            detail['time'] = time.strftime('%H:%M')
            detail['message'] = format_markdown(detail['message'])
            data['details'].append(detail)
        time += datetime.timedelta(minutes=minutes)
        data['details'].append(dict(
            time=time.strftime('%H:%M'), message='<p>Ende</p>',
            responsible=''))
        # Commitments
        data['commitments'] = []
        for commitment in self.context.commitments:
            data['commitments'].append(self.get_json(commitment))
        # Attachments
        data['attachments'] = bool(self.context.attachments)
        if self.context.day >= datetime.datetime.now():
            data['type'] = 'AGENDA'
        else:
            data['type'] = 'PROTOKOLL'
        compiler = pybars.Compiler()
        template = compiler.compile(TEMPLATE)
        html = "".join(template(data))
        pdf = BytesIO()
        xhtml2pdf.pisa.CreatePDF(html, dest=pdf)
        pdf.seek(0)

        output = PdfFileWriter()
        append_pdf(PdfFileReader(pdf, strict=False), output)

        for attachment in self.context.attachments:
            if attachment.mimetype in ('application/pdf',):
                pdf = attachment.data
            else:
                pdf = img2pdf.convert(attachment.data)
            append_pdf(PdfFileReader(BytesIO(pdf), strict=False), output)

        result = BytesIO()
        output.write(result)
        result.seek(0)
        return result



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import datetime
import transaction


def setUp():
    from sw.allotmentclub import Depot, User
    user = User.create(username='hans')
    Depot.create(date=datetime.datetime(2014, 11, 27, 7, 21, 45),
                 size=15, data=b'GIF89a????!?,D;', mimetype='image/gif',
                 name='test.gif', user=user)
    transaction.commit()


def test_displays_list_of_depots(browser):
    setUp()
    browser.login()
    browser.open('http://localhost/depots')
    expected = [1, 'test.gif', 'GIF', '15.00 B', '27.11.2014 07:21', 'hans']
    assert expected in browser.json_result


def test_depot_can_be_added_via_json_view(browser):
    from sw.allotmentclub import Depot
    browser.login()
    browser._upload('http://localhost/depots/add')
    assert 'success' == browser.json['status']
    assert 'Transparent.gif' == Depot.query().one().name


def test_depot_can_be_edited_via_json_view(browser):
    from sw.allotmentclub import Depot
    setUp()
    browser.login()
    browser._upload('http://localhost/depots/1/edit')
    now = datetime.datetime.now()
    assert 'success' == browser.json['status']
    assert 1 == len(Depot.query().all())
    browser.open('http://localhost/depots')
    assert [1, 'Transparent.gif', 'GIF', '42.00 B',
            now.strftime('%d.%m.%Y %H:%M'), 'admin'] == browser.json_result[0]


def test_depot_can_not_edit_file_if_mimetype_is_different(browser):
    from sw.allotmentclub import Depot
    setUp()
    depot = Depot.get(1)
    depot.mimetype = 'image/png'
    transaction.savepoint()
    browser.login()
    browser._upload('http://localhost/depots/1/edit')
    assert 'error' == browser.json['status']
    assert (
        'Kann nur Dateien vom gleichen Typ bearbeiten. '
        'alt: image/png, neu: image/gif' == browser.json['message'])



"""Replace electric meter for Steinmüller (Zählerwechsel)

Revision ID: 897a3181401e
Revises: 9f65f923772f
Create Date: 2020-08-24 07:29:55.519846

"""
from alembic import op
import risclog.sqlalchemy

# revision identifiers, used by Alembic.
revision = '897a3181401e'
down_revision = '9f65f923772f'

OLD_NUMBER = '42418392'
OLD_VALUE = 28610
NEW_NUMBER = '30499547'
NEW_VALUE = 0
YEAR = 2020


def upgrade():
    from sw.allotmentclub import ElectricMeter
    import transaction

    db = risclog.sqlalchemy.db.get_database()
    db._engines['portal'] = dict(
        engine=op.get_bind().engine,
        alembic_location="sw.allotmentclub:portal",
    )

    old = (ElectricMeter.query()
           .filter(ElectricMeter.number == OLD_NUMBER)
           .one())
    old.replace(old_value=OLD_VALUE, new_number=NEW_NUMBER,
                new_value=NEW_VALUE)
    transaction.commit()


def downgrade():
    from sw.allotmentclub import ElectricMeter
    import transaction

    db = risclog.sqlalchemy.db.get_database()
    db._engines['portal'] = dict(
        engine=op.get_bind().engine,
        alembic_location="sw.allotmentclub:portal",
    )

    to_delete = (ElectricMeter.query()
                 .filter(ElectricMeter.number == NEW_NUMBER)
                 .one())
    to_delete.get_value(YEAR).delete()
    to_delete.delete()

    old = (ElectricMeter.query()
           .filter(ElectricMeter.number == OLD_NUMBER)
           .one())
    old.get_value(YEAR-1).delete()
    transaction.commit()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .depot import MAX_FILE_SIZE
from .model import Object, Member
from sqlalchemy import (
    Column, Integer, String, ForeignKey, DateTime, Boolean, sql, LargeBinary)
import sqlalchemy.orm


class Keylist(Object):
    """Ein Eintrag im Schlüsselbuch."""

    subject = Column(String)


class Key(Object):
    """Ein Schlüssel."""

    keylist_id = Column(
        Integer, ForeignKey(Keylist.id, ondelete='cascade'), nullable=False)
    keylist = sqlalchemy.orm.relation(
        'Keylist', uselist=False, backref='keys', cascade='all')

    serial = Column(String)

    member_id = Column(Integer, ForeignKey(Member.id), nullable=True)
    member = sqlalchemy.orm.relation('Member', uselist=False, backref='keys')
    rent = Column(DateTime, nullable=True)
    note = Column(String)
    lost = Column(Boolean, default=sql.false())


class KeylistAttachment(Object):
    """Eine Anlage zum Schlüsselbuch."""

    keylist_id = Column(
        Integer, ForeignKey(Keylist.id, ondelete='cascade'), nullable=False)
    keylist = sqlalchemy.orm.relation(
        'Keylist', uselist=False, backref='attachments', cascade='all')
    name = Column(String(100), default=u'')
    mimetype = Column(String(30), default=u'')
    size = Column(String(20), default=u'')
    data = Column(LargeBinary(MAX_FILE_SIZE))



"""Update member data.

Revision ID: 2d287284c72e
Revises: 50035156da61
Create Date: 2015-09-14 10:34:10.825636

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '2d287284c72e'
down_revision = '50035156da61'


def upgrade():
    op.add_column(
        'energyvalue', sa.Column('to_pay', sa.Integer(), nullable=True))
    op.execute("UPDATE energyvalue SET to_pay = 0")
    op.execute("""
        INSERT into electricmeter
               (id, allotment_id, number, electric_power, disconnected)
            VALUES
               (124, 197, 0, False, False)""")
    op.execute("""
        INSERT into energyvalue (electric_meter_id, year, value, member_id,
               advance_pay, fee, price, usage, whole_price, discounted, to_pay)
            VALUES (124, 2014, 0, 188, 0, 0, 0, 0, 0, true, 0)""")
    op.execute("""
        INSERT into bookingkind
               (title, year, value_per_member, banking_account_id)
            VALUES ('Mitgliedsbeitrag', 2016, 750000, 4)""")
    op.execute("""
        INSERT into bookingkind (title, year, banking_account_id)
            VALUES ('Energieabschlag I', 2016, 4)""")
    op.execute("""
        INSERT into bookingkind (title, year, banking_account_id)
            VALUES ('Energieabschlag II', 2016, 4)""")
    op.execute("""
        INSERT into bookingkind (title, year, banking_account_id)
            VALUES ('Energieabrechnung', 2016, 4)""")
    op.execute("DELETE FROM booking WHERE id in (404, 405, 406, 407)")
    op.execute(
        "UPDATE member SET email = 'boehme.wsf@online.de' WHERE id = 126")
    op.execute("""
        UPDATE energyprice
            SET price = 2786,
                normal_fee = 67000,
                power_fee = 201100
            WHERE year = 2015""")


def downgrade():
    op.drop_column('energyvalue', 'to_pay')
